import os
from typing import List

import appdirs
from beartype import beartype
from plumbum import local

from pyrovelocity.logging import configure_logging


logger = configure_logging(__name__)


def is_aria2c_installed() -> bool:
    """Check if aria2c is installed."""
    aria2c = local.get("aria2c", None)
    return aria2c is not None


@beartype
def download_with_aria2c(urls: List[str], cache_dir: str):
    """
    Use aria2c to download files in parallel.
    """
    aria2c = local["aria2c"]
    aria2c_cmd = aria2c[
        "-x",
        "16",
        "-s",
        "16",
        "-j",
        "16",
        "-k",
        "1M",
        "-d",
        cache_dir,
        "--continue",
    ]

    for url in urls:
        try:
            logger.info(f"Downloading {url} to {cache_dir}")
            aria2c_cmd(url)
        except Exception as e:
            logger.error(f"Failed to download {url}: {e}")


@beartype
def download(urls: List[str], subdir: str = "pyrovelocity"):
    """
    Download all URLs in the provided list `urls` in parallel, supporting resume.
    Uses aria2c if available, otherwise logs an error.

    Args:
        urls: List of URLs to download.

    Example:
        >>> # xdoctest: +SKIP
        >>> from pyrovelocity.io.fetch import download
        >>> urls = [
        ...     "https://ftp.ensembl.org/pub/release-110/fasta/homo_sapiens/dna/Homo_sapiens.GRCh38.dna.chromosome.21.fa.gz",
        ...     "https://ftp.ensembl.org/pub/release-110/fasta/homo_sapiens/dna/Homo_sapiens.GRCh38.dna.chromosome.22.fa.gz",
        ... ]
        >>> download(urls)

        >>> # xdoctest: +SKIP
        >>> from pyrovelocity.io.fetch import download
        >>> urls = [
        ...     "https://ftp.ensembl.org/pub/release-110/fasta/homo_sapiens/dna/Homo_sapiens.GRCh38.dna.primary_assembly.fa.gz",
        ...     "https://ftp.ensembl.org/pub/release-110/fasta/mus_musculus/dna/Mus_musculus.GRCm39.dna.primary_assembly.fa.gz",
        ... ]
        >>> download(urls)
    """
    if not is_aria2c_installed():
        logger.error(
            "aria2c is not installed. Please install aria2c to proceed with downloads."
        )
        return

    try:
        cache_dir = os.path.join(appdirs.user_cache_dir(), subdir)
        os.makedirs(cache_dir, exist_ok=True)
        download_with_aria2c(urls, cache_dir)
    except KeyboardInterrupt:
        logger.info(
            f"Discontinuing download of\n\t{urls} to\n\t{cache_dir}\n\n"
        )



"""Tests for `pyrovelocity.utils` module."""

import hypothesis
import numpy as np
import pytest
from anndata import AnnData
from hypothesis import given
from hypothesis import strategies as st

from pyrovelocity.utils import generate_sample_data


def test_load_utils():
    from pyrovelocity import utils

    print(utils.__file__)


n_obs_strategy = st.integers(min_value=5, max_value=100)
n_vars_strategy = st.integers(min_value=5, max_value=100)
rate_strategy = st.floats(min_value=0.01, max_value=3.0)
beta_gamma_strategy = st.tuples(
    st.floats(min_value=0.01, max_value=3.0),  # beta
    st.floats(min_value=0.01, max_value=3.0),  # gamma
).filter(lambda x: x[1] > x[0])  # gamma required to be larger than beta
noise_model_strategy = st.sampled_from(["iid", "normal"])
random_seed_strategy = st.randoms()


@hypothesis.settings(deadline=1500)
@given(
    n_obs=n_obs_strategy,
    n_vars=n_vars_strategy,
    alpha=rate_strategy,
    beta_gamma=beta_gamma_strategy,
    alpha_=rate_strategy,
    noise_model=noise_model_strategy,
    random_seed=random_seed_strategy,
)
def test_generate_sample_data(
    n_obs: int,
    n_vars: int,
    alpha: float,
    beta_gamma: tuple,
    alpha_: float,
    noise_model: str,
    random_seed: np.random.RandomState,
):
    beta, gamma = beta_gamma
    seed = random_seed.randint(0, 1000)

    adata = generate_sample_data(
        n_obs=n_obs,
        n_vars=n_vars,
        alpha=alpha,
        beta=beta,
        gamma=gamma,
        alpha_=alpha_,
        noise_model=noise_model,
        random_seed=seed,
    )

    assert isinstance(adata, AnnData)
    assert adata.shape == (n_obs, n_vars)
    assert "spliced" in adata.layers
    assert "unspliced" in adata.layers

    new_seed = seed
    while new_seed == seed:
        new_seed = random_seed.randint(0, 1000)

    # Check that different seeds produce different results
    adata2 = generate_sample_data(
        n_obs=n_obs,
        n_vars=n_vars,
        alpha=alpha,
        beta=beta,
        gamma=gamma,
        alpha_=alpha_,
        noise_model=noise_model,
        random_seed=new_seed,
    )

    assert not (adata.X == adata2.X).all()


@pytest.mark.parametrize("n_obs, n_vars", [(100, 12), (50, 10), (200, 20)])
@pytest.mark.parametrize("noise_model", ["iid", "gillespie", "normal"])
def test_generate_sample_data_dimensions(n_obs, n_vars, noise_model):
    adata = generate_sample_data(
        n_obs=n_obs, n_vars=n_vars, noise_model=noise_model, random_seed=98
    )
    assert adata.shape == (n_obs, n_vars)


def test_generate_sample_data_layers(default_sample_data):
    assert "spliced" in default_sample_data.layers
    assert "unspliced" in default_sample_data.layers


def test_generate_sample_data_reproducibility(default_sample_data):
    adata1 = default_sample_data
    adata2 = generate_sample_data(random_seed=98)
    assert (adata1.X == adata2.X).all()
    assert (adata1.layers["spliced"] == adata2.layers["spliced"]).all()
    assert (adata1.layers["unspliced"] == adata2.layers["unspliced"]).all()


def test_generate_sample_data_invalid_noise_model():
    with pytest.raises(
        ValueError,
        match="noise_model must be one of 'iid', 'gillespie', 'normal'",
    ):
        generate_sample_data(noise_model="wishful thinking")



import math
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import mlflow
import numpy as np
import pyro
import scipy
import torch
from pyro.infer import Trace_ELBO
from pyro.infer import TraceEnum_ELBO
from pyro.infer.autoguide.guides import AutoGuideList
from pyro.optim.clipped_adam import ClippedAdam
from pyro.optim.optim import PyroOptim
from scvi.dataloaders import DataSplitter
from scvi.model._utils import parse_device_args
from scvi.train import PyroTrainingPlan
from scvi.train import TrainRunner

from pyrovelocity.logging import configure_logging
from pyrovelocity.models._velocity_module import VelocityModule


logger = configure_logging(__name__)


class VelocityAdam(ClippedAdam):
    def step(self, closure: Optional[Callable] = None) -> Optional[Any]:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            group["lr"] *= group["lrd"]

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                grad.clamp_(-group["clip_norm"], group["clip_norm"])
                # freeze non-velocity gene gradient
                grad[grad.isnan()] = 0.0
                state = self.state[p]

                # initialize state
                if len(state) == 0:
                    state["step"] = 0
                    # compute exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(grad)
                    # compute exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(grad)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                if group["weight_decay"] != 0:
                    grad = grad.add(p.data, alpha=group["weight_decay"])

                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                denom = exp_avg_sq.sqrt().add_(group["eps"])

                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]
                step_size = (
                    group["lr"] * math.sqrt(bias_correction2) / bias_correction1
                )

                p.data.addcdiv_(exp_avg, denom, value=-step_size)
        return loss


def VelocityClippedAdam(optim_args: Dict[str, float]) -> PyroOptim:
    """
    Wraps :class:`pyro.optim.clipped_adam.ClippedAdam` with :class:`~pyro.optim.optim.PyroOptim`.
    """
    return PyroOptim(VelocityAdam, optim_args)


class EnumTrainingPlan(PyroTrainingPlan):
    def __init__(
        self,
        pyro_velocity: VelocityModule,
        optim: Optional[pyro.optim.PyroOptim] = None,
    ):
        super().__init__(
            pyro_velocity,
            TraceEnum_ELBO(strict_enumeration_warning=True),
            optim,
        )
        self.svi = pyro.infer.SVI(
            model=self.module.model,
            guide=self.module.guide,
            optim=self.optim,
            loss=self.loss_fn,
        )
        self.n_elem = self.module.num_genes * self.module.num_cells * 2
        self.training_step_outputs = []
        self.validation_step_outputs = []

    # previously required optimizer_idx argument is no longer used
    def training_step(self, batch, batch_idx):
        args, kwargs = self.module._get_fn_args_from_batch(batch)
        loss = self.svi.step(*args, **kwargs)

        self.training_step_outputs.append(
            {
                "train_step_loss": loss,
                "num_elem": args[0].shape[0] * args[0].shape[1] * 2,
            }
        )
        return {
            "train_step_loss": loss,
            "num_elem": args[0].shape[0] * args[0].shape[1] * 2,
        }

    def on_train_epoch_end(self):
        n_batch, elbo = 0, 0
        for tensors in self.training_step_outputs:
            elbo += tensors["train_step_loss"]
            n_batch += 1
        if n_batch > 0:
            self.log("elbo_train", elbo / n_batch, prog_bar=True, on_epoch=True)

        self.training_step_outputs.clear()

    def validation_step(self, batch, batch_idx):
        args, kwargs = self.module._get_fn_args_from_batch(batch)
        loss = self.svi.evaluate_loss(*args, **kwargs)
        return {
            "valid_step_loss": loss,
            "num_elem": args[0].shape[0] * args[0].shape[1] * 2,
        }

    def on_validation_epoch_end(self):
        n_batch, elbo = 0, 0
        for tensors in self.validation_step_outputs:
            elbo += tensors["valid_step_loss"]
            n_batch += 1
        if n_batch > 0:
            self.log(
                "elbo_validation", elbo / n_batch, prog_bar=True, on_epoch=True
            )
        self.validation_step_outputs.clear()


class VelocityTrainingMixin:
    def train(
        self,
        use_gpu: str = "auto",
        early_stopping: bool = False,
        seed: int = 99,
        lr: float = 1e-3,
        train_size: float = 0.995,
        valid_size: float = 0.005,
        batch_size: int = 256,
        max_epochs: int = 100,
        check_val_every_n_epoch: Optional[int] = 1,
        patience: int = 10,
        min_delta: float = 0.0,
        **kwargs,
    ):
        logger.info(
            f"\ntrain model:\n"
            f"\ttraining fraction: {train_size}\n"
            f"\tvalidation fraction: {valid_size}\n"
        )
        pyro.clear_param_store()
        pyro.set_rng_seed(seed)

        data_splitter = DataSplitter(
            self.adata_manager,
            train_size=train_size,
            validation_size=valid_size,
            batch_size=batch_size,
        )
        data_splitter.setup()

        training_plan = EnumTrainingPlan(
            self.module, VelocityClippedAdam({"lr": lr, "lrd": 0.9999})
        )
        runner = TrainRunner(
            self,
            training_plan=training_plan,
            data_splitter=data_splitter,
            check_val_every_n_epoch=check_val_every_n_epoch,
            max_epochs=max_epochs,
            accelerator=use_gpu,
            early_stopping=early_stopping,
            early_stopping_patience=patience,
            early_stopping_min_delta=min_delta,
            **kwargs,
        )
        return runner()

    def train_faster(
        self,
        use_gpu: str = "auto",
        seed: int = 99,
        lr: float = 0.01,
        max_epochs: int = 5000,
        log_every: int = 100,
        patient_init: int = 45,
        patient_improve: float = 0.001,
    ) -> List[float]:
        """
        Train model with all data loaded into accelerator memory. 20k vars by 2k
        obs will consume approximately 40GB of GPU memory.

        Args:
            use_gpu (str, optional): _description_. Defaults to "auto".
            seed (int, optional): _description_. Defaults to 99.
            lr (float, optional): _description_. Defaults to 0.01.
            max_epochs (int, optional): _description_. Defaults to 5000.
            log_every (int, optional): _description_. Defaults to 100.
            patient_init (int, optional): _description_. Defaults to 45.
            patient_improve (float, optional): _description_. Defaults to 0.001.

        Returns:
            List[float]: _description_
        """

        logger.info(
            "train model in a single batch with all data loaded into accelerator memory"
        )

        _accelerator, _devices, device = parse_device_args(
            accelerator=use_gpu, return_device="torch"
        )
        logger.info(
            f"\nLoading model with:\n"
            f"\taccelerator: {_accelerator}\n"
            f"\tdevices: {_devices}\n"
            f"\tdevice: {device}\n\n"
        )

        pyro.clear_param_store()
        pyro.set_rng_seed(seed)
        pyro.enable_validation(True)
        optim = VelocityClippedAdam({"lr": lr, "lrd": 0.1 ** (1 / max_epochs)})
        self.module._model = self.module._model.to(device)
        # print("TraceEnum")
        svi = pyro.infer.SVI(
            self.module._model,
            self.module._guide,
            optim,
            Trace_ELBO(strict_enumeration_warning=True),
        )

        normalizer = self.adata.shape[0] * self.adata.shape[1] * 2
        u = torch.tensor(
            np.array(
                self.adata.layers["raw_unspliced"].toarray(), dtype="float32"
            )
            if scipy.sparse.issparse(self.adata.layers["raw_unspliced"])
            else self.adata.layers["raw_unspliced"],
            dtype=torch.float32,
        ).to(device)
        s = torch.tensor(
            np.array(
                self.adata.layers["raw_spliced"].toarray(), dtype="float32"
            )
            if scipy.sparse.issparse(self.adata.layers["raw_spliced"])
            else self.adata.layers["raw_spliced"],
            dtype=torch.float32,
        ).to(device)

        epsilon = 1e-6

        log_u_library_size = np.log(
            self.adata.obs.u_lib_size_raw.astype(float) + epsilon
        )
        log_s_library_size = np.log(
            self.adata.obs.s_lib_size_raw.astype(float) + epsilon
        )
        u_library = torch.tensor(
            np.array(log_u_library_size, dtype="float32"),
            dtype=torch.float32,
        ).to(device)
        s_library = torch.tensor(
            np.array(log_s_library_size, dtype="float32"),
            dtype=torch.float32,
        ).to(device)
        u_library_mean = (
            torch.tensor(
                np.mean(log_u_library_size),
                dtype=torch.float32,
            )
            .expand(u_library.shape)
            .to(device)
        )
        s_library_mean = (
            torch.tensor(
                np.mean(log_s_library_size),
                dtype=torch.float32,
            )
            .expand(u_library.shape)
            .to(device)
        )
        u_library_scale = (
            torch.tensor(
                np.std(log_u_library_size),
                dtype=torch.float32,
            )
            .expand(u_library.shape)
            .to(device)
        )
        s_library_scale = (
            torch.tensor(
                np.std(log_s_library_size),
                dtype=torch.float32,
            )
            .expand(u_library.shape)
            .to(device)
        )

        # print(u_library_scale.shape)
        # print(u.shape)
        # print(u_library.shape)
        if "pyro_cell_state" in self.adata.obs.columns:
            cell_state = torch.tensor(
                np.array(self.adata.obs.pyro_cell_state, dtype="float32"),
                dtype=torch.float32,
            ).to(device)
        else:
            cell_state = None

        losses = []
        patience = patient_init
        for step in range(max_epochs):
            if cell_state is None:
                elbos = (
                    svi.step(
                        u,
                        s,
                        u_library.reshape(-1, 1),
                        s_library.reshape(-1, 1),
                        u_library_mean.reshape(-1, 1),
                        s_library_mean.reshape(-1, 1),
                        u_library_scale.reshape(-1, 1),
                        s_library_scale.reshape(-1, 1),
                        None,
                        None,
                    )
                    / normalizer
                )
            else:
                elbos = (
                    svi.step(
                        u,
                        s,
                        u_library.reshape(-1, 1),
                        s_library.reshape(-1, 1),
                        u_library_mean.reshape(-1, 1),
                        s_library_mean.reshape(-1, 1),
                        u_library_scale.reshape(-1, 1),
                        s_library_scale.reshape(-1, 1),
                        None,
                        cell_state.reshape(-1, 1),
                    )
                    / normalizer
                )
            if (step == 0) or (
                ((step + 1) % log_every == 0) and ((step + 1) < max_epochs)
            ):
                mlflow.log_metric("-ELBO", -elbos, step=step + 1)
                logger.info(
                    f"step {step + 1: >4d} loss = {elbos:0.6g} patience = {patience}"
                )
            if step > log_every:
                if (losses[-1] - elbos) < losses[-1] * patient_improve:
                    patience -= 1
                else:
                    patience = patient_init
            if patience <= 0:
                break
            losses.append(elbos)
        mlflow.log_metric("-ELBO", -elbos, step=step + 1)
        mlflow.log_metric("real_epochs", step + 1)
        logger.info(
            f"step {step + 1: >4d} loss = {elbos:0.6g} patience = {patience}"
        )
        return losses

    def train_faster_with_batch(
        self,
        use_gpu: str = "auto",
        seed: int = 99,
        lr: float = 1e-2,
        max_epochs: int = 5000,
        log_every: int = 100,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = None,
        new_valid_guide: Optional[AutoGuideList] = None,
        patient_init: int = 45,
        patient_improve: float = 0.0,
        elbo_name: str = "-ELBO",
    ):
        logger.info("train model in batches of size {batch_size}")

        _accelerator, _devices, device = parse_device_args(
            accelerator=use_gpu, return_device="torch"
        )
        logger.info(
            f"\nLoading model with:\n"
            f"\taccelerator: {_accelerator}\n"
            f"\tdevices: {_devices}\n"
            f"\tdevice: {device}\n\n"
        )

        pyro.clear_param_store()
        pyro.set_rng_seed(seed)
        pyro.enable_validation(True)

        adata = self._validate_anndata(self.adata)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size
        )
        optim = VelocityClippedAdam({"lr": lr, "lrd": 0.1 ** (1 / max_epochs)})
        self.module._model = self.module._model.to(device)

        if new_valid_guide is None:
            svi = pyro.infer.SVI(
                self.module._model,
                self.module._guide,
                optim,
                Trace_ELBO(strict_enumeration_warning=True),
            )
        else:
            svi = pyro.infer.SVI(
                self.module._model,
                new_valid_guide,
                optim,
                Trace_ELBO(strict_enumeration_warning=True),
            )

        losses = []
        patience = patient_init
        for step in range(max_epochs):
            n_batch = 0
            elbos = 0
            for tensor_dict in scdl:
                args, kwargs = self.module._get_fn_args_from_batch(tensor_dict)
                args = [a.to(device) if a is not None else a for a in args]
                loss = svi.step(*args, **kwargs)
                elbos += loss
                n_batch += 1
            elbos = elbos / n_batch
            if (step == 0) or (
                ((step + 1) % log_every == 0) and ((step + 1) < max_epochs)
            ):
                mlflow.log_metric("-ELBO", -elbos, step=step + 1)
                logger.info(
                    f"step {step + 1: >4d} loss = {elbos:0.6g} patience = {patience}"
                )
            if step > log_every:
                if (losses[-1] - elbos) < losses[-1] * patient_improve:
                    patience -= 1
                else:
                    patience = patient_init
            if patience <= 0:
                break
            losses.append(elbos)
        mlflow.log_metric("-ELBO", -elbos, step=step + 1)
        mlflow.log_metric("real_epochs", step + 1)
        logger.info(
            f"step {step: >4d} loss = {elbos:0.6g} patience = {patience}"
        )
        return losses



import logging
from typing import Tuple

import diffrax
import jax
import jax.numpy as jnp
from beartype import beartype
from beartype.typing import Callable
from diffrax import ODETerm
from diffrax import SaveAt
from diffrax import Solution
from diffrax import Tsit5
from diffrax import diffeqsolve
from jaxtyping import ArrayLike
from jaxtyping import Float
from jaxtyping import PyTree
from jaxtyping import jaxtyped

from pyrovelocity.logging import configure_logging


__all__ = [
    "solve_transcription_splicing_model",
    "solve_transcription_splicing_model_analytical",
]

logging.getLogger("jax").setLevel(logging.ERROR)
logger = configure_logging(__name__)


TimeArrayType = Float[ArrayLike, "num_times"]
StateType = Float[ArrayLike, "num_vars"]
ParamsType = Tuple[float, ...]
ModelType = Callable[
    [
        PyTree[ArrayLike, "num_times"],
        PyTree[ArrayLike, "num_vars"],
        Tuple[float, ...],
    ],
    PyTree[ArrayLike, "num_vars"],
]


@jaxtyped(typechecker=beartype)
def dstate_dt_dimless(
    t: PyTree[ArrayLike, "num_times"],
    state: PyTree[ArrayLike, "num_vars"],
    params: PyTree[ArrayLike, "num_params"],
) -> PyTree[ArrayLike, "num_vars"]:
    """
    Simulate the transcription-splicing model.

    Args:
        t (float): time
        state (StateType): state vector
        params (ParamsType): scalar parameter

    Returns:
        PyTree[ArrayLike, "num_vars"]:
            time derivative of the state vector like StateType
    """
    u_star, s_star = state
    (gamma_star,) = params
    du_star_dt = 1 - u_star
    ds_star_dt = u_star - gamma_star * s_star
    return jnp.stack([du_star_dt, ds_star_dt])


@jaxtyped(typechecker=beartype)
def dstate_dt_dimensioned(
    t: PyTree[ArrayLike, "num_times"],
    state: PyTree[ArrayLike, "num_vars"],
    params: PyTree[ArrayLike, "num_params"],
) -> PyTree[ArrayLike, "num_vars"]:
    u, s = state
    alpha, beta, gamma = params
    du_dt = alpha - beta * u
    ds_dt = beta * u - gamma * s
    return jnp.stack([du_dt, ds_dt])


@jaxtyped(typechecker=beartype)
def solve_transcription_splicing_model_impl(
    ts: TimeArrayType,
    initial_state: StateType,
    params: ParamsType,
    time_step: float = 0.01,
    model: ModelType = dstate_dt_dimless,
) -> Solution:
    """
    Solve the transcription-splicing model.

    Args:
        ts (TimeArrayType): array of time points
        initial_state (StateType): initial state vector
        params (ParamsType): Tuple of scalar parameter(s).
        time_step (float): time step for the solver.
        model (ModelType): model function.

    Returns:
        Solution: diffrax Solution object.
    """
    term = ODETerm(model)
    t0 = ts[0]
    t1 = ts[-1]
    saveat = SaveAt(ts=ts)

    solution = diffeqsolve(
        term,
        solver=Tsit5(),
        t0=t0,
        t1=t1,
        dt0=time_step,
        y0=initial_state,
        args=params,
        saveat=saveat,
    )

    return solution


def solve_transcription_splicing_model(
    ts: TimeArrayType,
    initial_state: StateType,
    params: ParamsType,
    time_step: float = 0.01,
    model: ModelType = dstate_dt_dimless,
) -> Solution:
    """
    Solve the transcription-splicing model.

    This is a mirror of the `solve_transcription_splicing_model_impl` function
    with type-checking disabled for JIT compilation. This function should be
    removed when JIT compilation is enabled for the type-checked function.

    Args:
        ts (TimeArrayType): array of time points
        initial_state (StateType): initial state vector
        params (ParamsType): Tuple of scalar parameter(s).
        time_step (float): time step for the solver.
        model (ModelType): model function.

    Returns:
        Solution: diffrax Solution object.

    Examples:
        >>> from pyrovelocity.models import solve_transcription_splicing_model
        >>> from jax import numpy as jnp
        >>> ts0 = jnp.linspace(0, 4, 40)
        >>> ts1 = jnp.linspace(4 + (10 - 4) / 20, 10, 20)
        >>> ts = jnp.concatenate([ts0, ts1])
        >>> initial_state = jnp.array([0.1, 0.1])
        >>> params = (0.99,)
        >>> solution = solve_transcription_splicing_model(
        ...     ts,
        ...     initial_state,
        ...     params,
        >>> )
    """
    term = ODETerm(model)
    t0 = ts[0]
    t1 = ts[-1]
    saveat = SaveAt(ts=ts)

    solution = diffeqsolve(
        term,
        solver=Tsit5(),
        t0=t0,
        t1=t1,
        dt0=time_step,
        y0=initial_state,
        args=params,
        saveat=saveat,
    )

    return solution


# TODO: enable JIT compilation of the type-checked solve function.
solve_transcription_splicing_model = jax.jit(
    # solve_transcription_splicing_model_impl,
    solve_transcription_splicing_model,
    static_argnames=["model"],
)


@jaxtyped(typechecker=beartype)
def calculate_xi(
    initial_u_star: Float[ArrayLike, ""],
    gamma_star: Float[ArrayLike, ""],
):
    """
    Calculate the Xi parameter for the analytical solution.

    Args:
        initial_u_star (float): The initial value of u_star.
        gamma_star (float): The gamma_star parameter.

    Returns:
        A float representing the value of Xi.
    """
    return (initial_u_star - 1) / (gamma_star - 1)


@jaxtyped(typechecker=beartype)
def analytical_solution_dstate_dt_dimless(
    ts: TimeArrayType,
    initial_state: StateType,
    params: ParamsType,
) -> PyTree[ArrayLike, "num_vars"]:
    """
    Compute the analytical solution for the dimensionless transcription-splicing
    model.

    Args:
        ts (TimeArrayType): Array of time points. initial_state (StateType):
        Initial state vector [u_star_0, s_star_0]. params (ParamsType): Tuple
        containing a single parameter (gamma_star).

    Returns:
        PyTree[ArrayLike, "num_vars"]:
            A PyTree containing the solution arrays for u_star and s_star.
    """
    (gamma_star,) = params
    u_star_0, s_star_0 = initial_state
    t_star_0 = ts[0]
    elapsed_ts = ts - t_star_0

    u_star = 1 + (u_star_0 - 1) * jnp.exp(-elapsed_ts)

    if jnp.isclose(gamma_star, 1.0):
        s_star = (
            1
            + (s_star_0 - 1) * jnp.exp(-elapsed_ts)
            + (u_star_0 - 1) * elapsed_ts * jnp.exp(-elapsed_ts)
        )
    else:
        xi = calculate_xi(u_star_0, gamma_star)

        s_star = (1 / gamma_star) + (
            (s_star_0 - xi - (1 / gamma_star))
            * jnp.exp(-gamma_star * elapsed_ts)
            + xi * jnp.exp(-elapsed_ts)
        )

    return jnp.stack([u_star, s_star]).T


@jaxtyped(typechecker=beartype)
def solve_transcription_splicing_model_analytical(
    ts: TimeArrayType,
    initial_state: StateType,
    params: ParamsType,
) -> Solution:
    """
    Computes the solutions using the analytical approach and wraps the result in
    a Diffrax Solution object.

    Args:
        ts: The time points at which to evaluate the solution.
        initial_state: The initial conditions for the system.
        params: The parameters of the model.

    Returns:
        Solution: A Diffrax Solution object containing the analytical solutions.

    Examples:
        >>> from pyrovelocity.models import solve_transcription_splicing_model_analytical
        >>> from jax import numpy as jnp
        >>> ts0 = jnp.linspace(0, 4, 40)
        >>> ts1 = jnp.linspace(4 + (10 - 4) / 20, 10, 20)
        >>> ts = jnp.concatenate([ts0, ts1])
        >>> initial_state = jnp.array([0.1, 0.1])
        >>> params = (0.99,)
        >>> analytical_solution = solve_transcription_splicing_model_analytical(
        ...     ts,
        ...     initial_state,
        ...     params,
        >>> )
    """

    ys = analytical_solution_dstate_dt_dimless(
        ts,
        initial_state,
        params,
    )

    solution = Solution(
        t0=ts[0],
        t1=ts[-1],
        ts=ts,
        ys=ys,
        interpolation=None,
        stats={},
        result=diffrax.RESULTS.successful,
        solver_state=None,
        controller_state=None,
        made_jump=None,
    )

    return solution



import shutil

import jax.numpy as jnp
import matplotlib.pyplot as plt
from beartype import beartype
from beartype.typing import List
from diffrax import Solution
from matplotlib import colors

__all__ = [
    "plot_deterministic_simulation_phase_portrait",
    "plot_deterministic_simulation_trajectories",
]


@beartype
def plot_deterministic_simulation_trajectories(
    solution: Solution,
    state_labels: List[str] = [r"$u^{\ast}$", r"$s^{\ast}$"],
    xlabel: str = r"$t^{\ast}$",
    ylabel: str = r"fraction of $u^{\ast}_{ss}$",
    title_prefix: str = "",
    colormap_name: str = "cividis",
):
    times = solution.ts
    logarithmizable_times = jnp.where(times <= 1e-4, 1e-2, times)
    unspliced, spliced = solution.ys[:, 0], solution.ys[:, 1]
    logarithmizable_unspliced = jnp.where(
        jnp.abs(unspliced) <= 3e-8, 1e-8, jnp.abs(unspliced)
    )
    logarithmizable_spliced = jnp.where(
        jnp.abs(spliced) <= 3e-8, 1e-8, jnp.abs(spliced)
    )

    with plt.style.context(["pyrovelocity.styles.common"]):
        if not shutil.which("latex"):
            plt.rc(
                "text",
                usetex=False,
            )
        fig, axes = plt.subplots(1, 2, figsize=(14, 5))

        cmap = plt.get_cmap(colormap_name)
        color_brown = cmap(0.05)
        color_blue_green = cmap(0.95)

        # linear scale
        axes[0].plot(
            times,
            unspliced,
            "o-",
            label=state_labels[0],
            color=color_brown,
            markersize=5,
            linewidth=1.5,
            alpha=0.5,
        )
        axes[0].plot(
            times,
            spliced,
            "o-",
            label=state_labels[1],
            color=color_blue_green,
            markersize=5,
            linewidth=1.5,
            alpha=0.5,
        )
        axes[0].set_xlabel(xlabel)
        axes[0].set_ylabel(ylabel)
        axes[0].set_title(
            f"{title_prefix} Temporal Trajectories (Linear Scale)"
        )
        axes[0].legend()

        # log scale
        axes[1].plot(
            logarithmizable_times,
            logarithmizable_unspliced,
            "o-",
            label=state_labels[0],
            color=color_brown,
            markersize=5,
            linewidth=1.5,
            alpha=0.5,
        )
        axes[1].plot(
            logarithmizable_times,
            logarithmizable_spliced,
            "o-",
            label=state_labels[1],
            color=color_blue_green,
            markersize=5,
            linewidth=1.5,
            alpha=0.5,
        )
        axes[1].set_xscale("log")
        axes[1].set_yscale("log")
        axes[1].set_xlabel(xlabel)
        axes[1].set_ylabel(ylabel)
        axes[1].set_title(f"{title_prefix} Temporal Trajectories (Log Scale)")
        axes[1].legend()

        plt.tight_layout()
        plt.show()


@beartype
def plot_deterministic_simulation_phase_portrait(
    solution: Solution,
    xlabel: str = r"$s^{\ast}$",
    ylabel: str = r"$u^{\ast}$",
    zlabel: str = r"$t^{\ast}$",
    title_prefix: str = "",
    colormap_name: str = "cividis",
):
    times = solution.ts
    logarithmizable_times = jnp.where(times <= 1e-4, 1e-2, times)

    unspliced, spliced = solution.ys[:, 0], solution.ys[:, 1]

    with plt.style.context(["pyrovelocity.styles.common"]):
        if not shutil.which("latex"):
            plt.rc(
                "text",
                usetex=False,
            )
        fig, axes = plt.subplots(1, 2, figsize=(14, 5))

        # linear scale
        sc_linear = axes[0].scatter(
            spliced,
            unspliced,
            c=times,
            cmap=colormap_name,
            label="Phase Portrait",
        )
        axes[0].set_xlabel(xlabel)
        axes[0].set_ylabel(ylabel)
        axes[0].set_title(f"{title_prefix} Phase Portrait (Linear Scale)")
        plt.colorbar(sc_linear, ax=axes[0], label=zlabel)

        # log scale
        sc_log = axes[1].scatter(
            spliced,
            unspliced,
            c=logarithmizable_times,
            cmap=colormap_name,
            norm=colors.LogNorm(),
            label="Phase Portrait",
        )
        axes[1].set_xscale("log")
        axes[1].set_yscale("log")
        axes[1].set_xlabel(xlabel)
        axes[1].set_ylabel(ylabel)
        axes[1].set_title(f"{title_prefix} Phase Portrait (Log Scale)")
        plt.colorbar(sc_log, ax=axes[1], label=zlabel)
        plt.tight_layout()
        plt.show()



import anndata
import numpy as np
import pytest
import scvelo as scv
from pyrovelocity.metrics.criticality_index import (
    calculate_criticality_index,
)


@pytest.fixture(scope="module")
def adata_fixture():
    return scv.datasets.simulation(
        random_seed=0,
        n_obs=10,
        n_vars=3,
        alpha=5,
        beta=0.5,
        gamma=0.3,
        alpha_=0,
        switches=[1, 5, 10],
        noise_model="gillespie",
    )


def test_calculate_criticality_index(adata_fixture):
    criticality_index, _, _, _ = calculate_criticality_index(adata_fixture)
    assert isinstance(criticality_index, float)
    assert 0 <= criticality_index <= 1


def test_empty_anndata():
    empty_adata = anndata.AnnData(X=np.empty((0, 0)))
    with pytest.raises(ValueError):
        calculate_criticality_index(empty_adata)


def test_missing_unspliced_layer(adata_fixture):
    adata_no_unspliced = adata_fixture.copy()
    del adata_no_unspliced.layers["unspliced"]
    with pytest.raises(KeyError):
        calculate_criticality_index(adata_no_unspliced)


def test_missing_spliced_layer(adata_fixture):
    adata_no_spliced = adata_fixture.copy()
    del adata_no_spliced.layers["spliced"]
    with pytest.raises(KeyError):
        calculate_criticality_index(adata_no_spliced)



import json
import os
import uuid
from pathlib import Path
from typing import Dict
from typing import Optional
from typing import Tuple

import matplotlib.pyplot as plt
import mlflow
import numpy as np
from anndata._core.anndata import AnnData
from beartype import beartype
from mlflow import MlflowClient
from numpy import ndarray
from scvi.model._utils import parse_device_args

from pyrovelocity.io.compressedpickle import CompressedPickle
from pyrovelocity.logging import configure_logging
from pyrovelocity.models import PyroVelocity
from pyrovelocity.tasks.data import load_anndata_from_path
from pyrovelocity.utils import print_anndata


logger = configure_logging(__name__)


@beartype
def train_dataset(
    adata: str | AnnData,
    data_set_name: str = "simulated",
    model_identifier: str = "model2",
    guide_type: str = "auto",
    model_type: str = "auto",
    batch_size: int = -1,
    use_gpu: str = "auto",
    likelihood: str = "Poisson",
    num_samples: int = 30,
    log_every: int = 100,
    patient_improve: float = 1e-4,
    patient_init: int = 45,
    seed: int = 99,
    learning_rate: float = 0.01,
    max_epochs: int = 3000,
    include_prior: bool = True,
    library_size: bool = True,
    offset: bool = True,
    input_type: str = "raw",
    cell_specific_kinetics: Optional[str] = None,
    kinetics_num: int = 2,
    force: bool = False,
) -> Tuple[str, str, Path, Path, Path, Path, Path, Path]:
    """
    Loads processed data, trains model, and saves model and posterior samples.

    Inputs:
        data/processed/{data_set_name}_processed.h5ad

    Outputs:
        models/{data_model}/
        ├── trained.h5ad
        ├── posterior_samples.pkl.zst
        └── model/
            ├── attr.pkl
            ├── model_params.pt
            ├── param_store_test.pt
            └── var_names.csv

    The file `models/{data_model}/pyrovelocity.pkl` used in
    `pyrovelocity.tasks.summarize.summarize_dataset` is produced by
    `pyrovelocity.tasks.postprocess.postprocess_dataset`.


    Args:
        adata (str | AnnData):
            Path to a file that can be read to an AnnData object or an AnnData object.
        data_set_name (str, optional): Name of the dataset. Default is "simulated".
        model_identifier (str, optional): Identifier for the model. Default is "model2".
        guide_type (str, optional):
            The type of guide function for the Pyro model. Default is "auto".
        model_type (str, optional): The type of Pyro model. Default is "auto".
        batch_size (int, optional):
            Batch size for training. Default is -1, which indicates using the full dataset.
        use_gpu (int, optional):
            Whether to use GPU for training. Default is 0, which indicates not using GPU.
        likelihood (str, optional):
            Likelihood function for the Pyro model. Default is "Poisson".
        num_samples (int, optional): Number of posterior samples. Default is 30.
        log_every (int, optional): Frequency of logging progress. Default is 100.
        patient_improve (float, optional):
            Minimum improvement in training loss for early stopping. Default is 5e-4.
        patient_init (int, optional):
            Number of initial training epochs before early stopping is enabled. Default is 30.
        seed (int, optional): Random seed for reproducibility. Default is 99.
        learning_rate (float, optional): Learning rate for the optimizer. Default is 0.01.
        max_epochs (int, optional): Maximum number of training epochs. Default is 3000.
        include_prior (bool, optional):
            Whether to include prior information in the model. Default is True.
        library_size (bool, optional): Whether to correct for library size. Default is True.
        offset (bool, optional): Whether to add an offset to the model. Default is False.
        input_type (str, optional): Type of input data. Default is "raw".
        cell_specific_kinetics (Optional[str], optional):
            Name of the attribute containing cell-specific kinetics information. Default is None.
        kinetics_num (int, optional): Number of kinetics parameters. Default is 2.
        force (bool, optional): Whether to overwrite existing files. Default is False.

    Returns:
        Tuple[str, str, Path, Path, Path, Path, Path, Path]:
            Tuple containing the name of the data model, the path to the model
            directory, the path to the trained data, the path to the model, the
            path to the posterior samples, the path to the metrics, the path to
            the run info, and the path to the loss plot.

    Examples:
        >>> train_dataset() # xdoctest: +SKIP
    """

    # load data
    data_model = f"{data_set_name}_{model_identifier}"
    data_model_path = Path(f"models/{data_model}")

    trained_data_path = data_model_path / "trained.h5ad"
    model_path = data_model_path / "model"
    posterior_samples_path = data_model_path / "posterior_samples.pkl.zst"
    metrics_path = data_model_path / "metrics.json"
    run_info_path = data_model_path / "run_info.json"
    loss_plot_path = data_model_path / "ELBO.png"

    logger.info(f"\n\nTraining: {data_model}\n\n")

    logger.info(
        f"\n\nVerifying existence of paths for:\n\n"
        f"  model data: {data_model_path}\n"
    )
    Path(data_model_path).mkdir(parents=True, exist_ok=True)

    # train model
    _accelerator, _devices, device = parse_device_args(
        accelerator=use_gpu, return_device="torch"
    )
    logger.info(
        f"Accelerator type specified as {use_gpu} resolves to:\n"
        f"\taccelerator: {_accelerator}\n"
        f"\tdevices: {_devices}\n"
        f"\tdevice: {device}\n"
    )

    if (
        os.path.isfile(trained_data_path)
        and os.path.exists(model_path)
        and os.path.isfile(posterior_samples_path)
        and not force
    ):
        logger.info(
            f"\n{trained_data_path}\n"
            f"{model_path}\n"
            f"{posterior_samples_path}\n"
            "all exist, set `force=True` to overwrite."
        )
        return (
            data_model,
            str(data_model_path),
            trained_data_path,
            model_path,
            posterior_samples_path,
            metrics_path,
            run_info_path,
            loss_plot_path,
        )
    else:
        logger.info(f"Training model: {data_model}")

        # UPDATE: v2.1.1 12/26/2022 autolog only supports pytorch lightning
        # mlflow.pytorch.autolog(log_every_n_epoch=200, log_models=False, silent=False)
        with mlflow.start_run(
            run_name=f"{data_model}-{uuid.uuid4().hex[:7]}"
        ) as run:
            mlflow.set_tag(
                "mlflow.runName", f"{data_model}-{run.info.run_id[:7]}"
            )
            print(f"Active run_id: {run.info.run_id}")
            mlflow.log_params(
                {
                    "data_set_name": data_set_name,
                    "model_identifier": model_identifier,
                    "guide_type": guide_type,
                    "model_type": model_type,
                    "batch_size": batch_size,
                }
            )

            adata, trained_model, posterior_samples = train_model(
                adata=adata,
                guide_type=guide_type,
                model_type=model_type,
                batch_size=batch_size,
                use_gpu=use_gpu,
                likelihood=likelihood,
                num_samples=num_samples,
                log_every=log_every,
                patient_improve=patient_improve,
                patient_init=patient_init,
                seed=seed,
                learning_rate=learning_rate,
                max_epochs=max_epochs,
                include_prior=include_prior,
                library_size=library_size,
                offset=offset,
                input_type=input_type,
                cell_specific_kinetics=cell_specific_kinetics,
                kinetics_num=kinetics_num,
                loss_plot_path=str(loss_plot_path),
            )

            logger.info("Data attributes after model training")

            run_id = run.info.run_id

        logger.info(f"Saving posterior samples: {posterior_samples_path}\n")
        CompressedPickle.save(
            posterior_samples_path,
            posterior_samples,
        )

        # save metrics
        r = mlflow.get_run(run_id)

        Path(metrics_path).write_text(json.dumps(r.data.metrics, indent=4))

        Path(run_info_path).write_text(
            json.dumps(r.to_dictionary()["info"], indent=4)
        )

        log_run_info(r)

        # postprocess
        if "pancreas" in data_model:
            logger.info("checking shared time")

            def check_shared_time(posterior_samples, adata):
                adata.obs["cell_time"] = (
                    posterior_samples["cell_time"].squeeze().mean(0)
                )
                adata.obs["1-Cytotrace"] = 1 - adata.obs["cytotrace"]

            check_shared_time(posterior_samples, adata)

        print_anndata(adata)

        # save checkpoints
        logger.info(f"Saving trained data: {trained_data_path}")
        adata.write(trained_data_path)

        logger.info(f"Saving model: {model_path}")
        trained_model.save_model(model_path, overwrite=True)

        logger.info(
            f"\nReturning paths to saved data:\n\n"
            f"{data_model_path}\n"
            f"{trained_data_path}\n"
            f"{model_path}\n"
            f"{posterior_samples_path}\n"
            f"{metrics_path}\n"
            f"{run_info_path}\n"
            f"{loss_plot_path}\n"
        )
        return (
            data_model,
            str(data_model_path),
            trained_data_path,
            model_path,
            posterior_samples_path,
            metrics_path,
            run_info_path,
            loss_plot_path,
        )


@beartype
def train_model(
    adata: str | AnnData,
    guide_type: str = "auto",
    model_type: str = "auto",
    batch_size: int = -1,
    use_gpu: str = "auto",
    likelihood: str = "Poisson",
    num_samples: int = 30,
    log_every: int = 100,
    patient_improve: float = 1e-4,
    patient_init: int = 45,
    seed: int = 99,
    learning_rate: float = 0.01,
    max_epochs: int = 3000,
    include_prior: bool = True,
    library_size: bool = True,
    offset: bool = True,
    input_type: str = "raw",
    cell_specific_kinetics: Optional[str] = None,
    kinetics_num: int = 2,
    loss_plot_path: str = "loss_plot.png",
) -> Tuple[AnnData, PyroVelocity, Dict[str, ndarray]]:
    """
    Train a PyroVelocity model to provide probabilistic estimates of RNA velocity
    for single-cell RNA sequencing data with quantified splice variants.

    Args:
        adata (str | AnnData): Path to a file that can be read to an AnnData object or an AnnData object.
        guide_type (str, optional): The type of guide function for the Pyro model. Default is "auto".
        model_type (str, optional): The type of Pyro model. Default is "auto".
        batch_size (int, optional): Batch size for training. Default is -1, which indicates using the full dataset.
        use_gpu (int, optional): Whether to use GPU for training. Default is 0, which indicates not using GPU.
        likelihood (str, optional): Likelihood function for the Pyro model. Default is "Poisson".
        num_samples (int, optional): Number of posterior samples. Default is 30.
        log_every (int, optional): Frequency of logging progress. Default is 100.
        patient_improve (float, optional): Minimum improvement in training loss for early stopping. Default is 5e-4.
        patient_init (int, optional): Number of initial training epochs before early stopping is enabled. Default is 30.
        seed (int, optional): Random seed for reproducibility. Default is 99.
        learning_rate (float, optional): Learning rate for the optimizer. Default is 0.01.
        max_epochs (int, optional): Maximum number of training epochs. Default is 3000.
        include_prior (bool, optional): Whether to include prior information in the model. Default is True.
        library_size (bool, optional): Whether to correct for library size. Default is True.
        offset (bool, optional): Whether to add an offset to the model. Default is False.
        input_type (str, optional): Type of input data. Default is "raw".
        cell_specific_kinetics (Optional[str], optional): Name of the attribute containing cell-specific kinetics information. Default is None.
        kinetics_num (int, optional): Number of kinetics parameters. Default is 2.
        loss_plot_path (str, optional): Path to save the loss plot. Default is "loss_plot.png".

        These arguments are deprecated:
        # svi_train: bool = False,
        # train_size: float = 1.0,
        # cell_state: str = "clusters",
        # svi_train (bool, optional): Whether to use Stochastic Variational Inference for training. Default is False.
        # train_size (float, optional): Proportion of data to be used for training. Default is 1.0.
        # cell_state (str, optional): Cell state attribute in the AnnData object. Default is "clusters".

    Returns:
        Tuple[PyroVelocity, Dict[str, ndarray]]: A tuple containing the trained PyroVelocity model and a dictionary of posterior samples.

    Examples:
        >>> from pyrovelocity.tasks.train import train_model
        >>> from pyrovelocity.utils import generate_sample_data
        >>> from pyrovelocity.tasks.preprocess import copy_raw_counts
        >>> tmp = getfixture("tmp_path")
        >>> loss_plot_path = str(tmp) + "/loss_plot_docs.png"
        >>> print(loss_plot_path)
        >>> adata = generate_sample_data(random_seed=99)
        >>> copy_raw_counts(adata)
        >>> _, model, posterior_samples = train_model(adata, use_gpu="auto", seed=99, max_epochs=200, loss_plot_path=loss_plot_path)
    """
    if isinstance(adata, str):
        adata = load_anndata_from_path(adata)

    logger.info(f"AnnData object prior to model training")
    print_anndata(adata)

    PyroVelocity.setup_anndata(adata)

    model = PyroVelocity(
        adata,
        likelihood=likelihood,
        model_type=model_type,
        guide_type=guide_type,
        correct_library_size=library_size,
        add_offset=offset,
        include_prior=include_prior,
        input_type=input_type,
        cell_specific_kinetics=cell_specific_kinetics,
        kinetics_num=kinetics_num,
    )

    if batch_size == -1:
        batch_size = adata.shape[0]

    if batch_size >= adata.shape[0]:
        losses = model.train_faster(
            max_epochs=max_epochs,
            lr=learning_rate,
            use_gpu=use_gpu,
            seed=seed,
            patient_improve=patient_improve,
            patient_init=patient_init,
            log_every=log_every,
        )
    else:
        losses = model.train_faster_with_batch(
            max_epochs=max_epochs,
            batch_size=batch_size,
            log_every=log_every,
            lr=learning_rate,
            use_gpu=use_gpu,
            seed=seed,
            patient_improve=patient_improve,
            patient_init=patient_init,
        )
    fig, ax = plt.subplots()
    fig.set_size_inches(2.5, 1.5)
    ax.scatter(
        np.arange(len(losses)),
        -np.array(losses),
        label="train",
        alpha=0.25,
    )
    set_loss_plot_axes(ax)
    posterior_samples = model.generate_posterior_samples(
        model.adata, num_samples=num_samples, batch_size=512
    )

    fig.savefig(loss_plot_path, facecolor="white", bbox_inches="tight")
    plt.close(fig)

    logger.info(f"AnnData object after model training")
    print_anndata(adata)
    return adata, model, posterior_samples


def set_loss_plot_axes(ax):
    ax.set_yscale("symlog")
    ax.set_xlabel("Epochs")
    ax.set_ylabel("-ELBO")


def log_run_info(r: mlflow.entities.run.Run) -> None:
    tags = {k: v for k, v in r.data.tags.items() if not k.startswith("mlflow.")}
    artifacts = [
        f.path for f in MlflowClient().list_artifacts(r.info.run_id, "model")
    ]
    logger.info(
        f"\nrun_id: {r.info.run_id}\n"
        f"artifacts: {artifacts}\n"
        f"params: {r.data.params}\n"
        f"metrics: {r.data.metrics}\n"
        f"tags: {tags}\n"
    )



import anndata
import numpy as np
from scipy.sparse import issparse


def get_clone_trajectory(
    adata,
    average_start_point=True,
    global_traj=True,
    times=[2, 4, 6],
    clone_num=None,
):
    if not average_start_point:
        adata.obsm["clone_vector_emb"] = np.zeros((adata.shape[0], 2))

    adatas = []
    clones = []
    centroids = []
    cen_clones = []
    print(adata.shape)
    adata.obs["clones"] = 0
    if "noWell" in adata.obs.columns:
        for w in adata.obs.Well.unique():
            adata_w = adata[adata.obs.Well == w]
            clone_adata_w = clone_adata[clone_adata.obs.Well == w]
            for j in range(clone_adata_w.shape[1]):
                adata_w.obs["clonei"] = 0
                # belongs to same clone
                adata_w.obs.loc[
                    clone_adata_w[:, j].X.toarray()[:, 0] >= 1, "clonei"
                ] = 1

                if not average_start_point:
                    for i in np.where(
                        (adata_w.obs.time == 2) & (adata_w.obs.clonei == 1)
                    )[0]:
                        next_time = np.where(
                            (adata_w.obs.time == 4) & (adata_w.obs.clonei == 1)
                        )[0]
                        adata_w.obsm["velocity_umap"][i] = (
                            adata_w.obsm["X_umap"][next_time].mean(axis=0)
                            - adata_w.obsm["X_umap"][i]
                        )
                    for i in np.where(
                        (adata_w.obs.time == 4) & (adata_w.obs.clonei == 1)
                    )[0]:
                        next_time = np.where(
                            (adata_w.obs.time == 6) & (adata_w.obs.clonei == 1)
                        )[0]
                        adata_w.obsm["velocity_umap"][i] = (
                            adata_w.obsm["X_umap"][next_time].mean(axis=0)
                            - adata_w.obsm["X_umap"][i]
                        )
                else:
                    time2 = np.where(
                        (adata_w.obs.time == 2) & (adata_w.obs.clonei == 1)
                    )[0]
                    time4 = np.where(
                        (adata_w.obs.time == 4) & (adata_w.obs.clonei == 1)
                    )[0]
                    time6 = np.where(
                        (adata_w.obs.time == 6) & (adata_w.obs.clonei == 1)
                    )[0]
                    if (
                        time2.shape[0] == 0
                        and time4.shape[0] == 0
                        and time6.shape[0] == 0
                    ):
                        continue
                    if (
                        time2.shape[0] > 0
                        and time4.shape[0] == 0
                        and time6.shape[0] > 0
                    ):
                        continue
                    adata_new = anndata.AnnData(
                        np.vstack(
                            [
                                adata_w[time2].X.toarray().mean(axis=0),
                                adata_w[time4].X.toarray().mean(axis=0),
                                adata_w[time6].X.toarray().mean(axis=0),
                            ]
                        ),
                        layers={
                            "spliced": np.vstack(
                                [
                                    adata_w[time2]
                                    .layers["spliced"]
                                    .toarray()
                                    .mean(axis=0),
                                    adata_w[time4]
                                    .layers["spliced"]
                                    .toarray()
                                    .mean(axis=0),
                                    adata_w[time6]
                                    .layers["spliced"]
                                    .toarray()
                                    .mean(axis=0),
                                ]
                            ),
                            "unspliced": np.vstack(
                                [
                                    adata_w[time2]
                                    .layers["unspliced"]
                                    .toarray()
                                    .mean(axis=0),
                                    adata_w[time4]
                                    .layers["unspliced"]
                                    .toarray()
                                    .mean(axis=0),
                                    adata_w[time6]
                                    .layers["unspliced"]
                                    .toarray()
                                    .mean(axis=0),
                                ]
                            ),
                        },
                        var=adata_w.var,
                    )

                    adata_new.obs.loc[:, "time"] = [2, 4, 6]
                    adata_new.obs.loc[:, "Cell type annotation"] = "Centroid"
                    print(adata_w[time6].obs.clonetype.unique())
                    print(adata_w[time6].obs)

                    adata_new.obs.loc[:, "clonetype"] = (
                        adata_w[time6].obs.clonetype.unique()
                    )  # use cell fate from last time point
                    adata_new.obs.loc[:, "clones"] = int(j)
                    if "Well" in adata_w[time6].obs.columns:
                        adata_new.obs.loc[:, "Well"] = adata_w[
                            time6
                        ].obs.Well.unique()

                    adata_new.obsm["X_umap"] = np.vstack(
                        [
                            adata_w[time2].obsm["X_umap"].mean(axis=0),
                            adata_w[time4].obsm["X_umap"].mean(axis=0),
                            adata_w[time6].obsm["X_umap"].mean(axis=0),
                        ]
                    )
                    adata_new.obsm["velocity_umap"] = np.vstack(
                        [
                            adata_w.obsm["X_umap"][time4].mean(axis=0)
                            - adata_w.obsm["X_umap"][time2].mean(axis=0),
                            adata_w.obsm["X_umap"][time6].mean(axis=0)
                            - adata_w.obsm["X_umap"][time4].mean(axis=0),
                            np.zeros(2),
                        ]
                    )
                    centroids.append(adata_new)
                    clone_new = anndata.AnnData(
                        np.vstack(
                            [
                                clone_adata_w[time2].X.toarray().mean(axis=0),
                                clone_adata_w[time4].X.toarray().mean(axis=0),
                                clone_adata_w[time6].X.toarray().mean(axis=0),
                            ]
                        ),
                        obs=adata_new.obs,
                    )
                    clone_new.var_names = clone_adata.var_names
                    clone_new.var = clone_adata.var
                    # print(clone_new.shape)
                    cen_clones.append(clone_new)

            adata_new = adata_w.concatenate(
                centroids[0].concatenate(centroids[1:]), join="outer"
            )
            clone_new = clone_adata_w.concatenate(
                cen_clones[0].concatenate(cen_clones[1:]), join="outer"
            )
            adatas.append(adata_new)
            clones.append(clone_new)
        return adatas[0].concatenate(adatas[1]), clones[0].concatenate(
            clones[1]
        )
    else:
        if clone_num is None:
            clone_num = adata.obsm["X_clone"].shape[1]
        for j in range(clone_num):
            print(j)
            adata.obs["clonei"] = 0
            # print('----------aa------')
            if issparse(adata.obsm["X_clone"]):
                adata.obs.loc[
                    adata.obsm["X_clone"].toarray()[:, j] >= 1, "clonei"
                ] = 1
            else:
                adata.obs.loc[adata.obsm["X_clone"][:, j] >= 1, "clonei"] = 1
            # print('----------bb------')

            if not average_start_point:
                for i in np.where(
                    (adata.obs.time == 2) & (adata.obs.clonei == 1)
                )[0]:
                    next_time = np.where(
                        (adata.obs.time == 4) & (adata.obs.clonei == 1)
                    )[0]
                    adata.obsm["velocity_umap"][i] = (
                        adata.obsm["X_umap"][next_time].mean(axis=0)
                        - adata.obsm["X_umap"][i]
                    )
                for i in np.where(
                    (adata.obs.time == 4) & (adata.obs.clonei == 1)
                )[0]:
                    next_time = np.where(
                        (adata.obs.time == 6) & (adata.obs.clonei == 1)
                    )[0]
                    adata.obsm["velocity_umap"][i] = (
                        adata.obsm["X_umap"][next_time].mean(axis=0)
                        - adata.obsm["X_umap"][i]
                    )
            else:
                if global_traj:
                    times_index = []
                    for t in times:
                        times_index.append(
                            np.where(
                                (adata.obs.time_info == t)
                                & (adata.obs.clonei == 1)
                            )[0]
                        )

                    consecutive_flag = np.array(
                        [int(time.shape[0] > 0) for time in times_index]
                    )
                    consecutive = np.diff(consecutive_flag)
                    if np.sum(consecutive_flag == 1) >= 2 and np.any(
                        consecutive == 0
                    ):  # Must be consecutive time points
                        # print('centroid:', consecutive, times_index)
                        adata_new = anndata.AnnData(
                            np.vstack(
                                [
                                    np.array(
                                        adata[time].X.mean(axis=0)
                                    ).squeeze()
                                    for time in times_index
                                    if time.shape[0] > 0
                                ]
                            ),
                            # layers={
                            #     "spliced": np.vstack(
                            #         [
                            #             np.array(
                            #                 adata[time]
                            #                 .layers["spliced"]
                            #                 .mean(axis=0)
                            #             )
                            #             for time in times_index
                            #             if time.shape[0] > 0
                            #         ]
                            #     ),
                            #     "unspliced": np.vstack(
                            #         [
                            #             np.array(
                            #                 adata[time]
                            #                 .layers["unspliced"]
                            #                 .mean(axis=0)
                            #             )
                            #             for time in times_index
                            #             if time.shape[0] > 0
                            #         ]
                            #     ),
                            # },
                            var=adata.var,
                        )
                        # print('----------cc------')
                        adata.obs.iloc[
                            np.hstack(
                                [
                                    time
                                    for time in times_index
                                    if time.shape[0] > 0
                                ]
                            ),
                            adata.obs.columns.get_loc("clones"),
                        ] = int(j)
                        adata_new.obs.loc[:, "time"] = [
                            t
                            for t, time in zip([2, 4, 6], times_index)
                            if time.shape[0] > 0
                        ]
                        adata_new.obs.loc[:, "clones"] = int(j)
                        adata_new.obs.loc[:, "state_info"] = "Centroid"
                        adata_new.obsm["X_emb"] = np.vstack(
                            [
                                adata[time].obsm["X_emb"].mean(axis=0)
                                for time in times_index
                                if time.shape[0] > 0
                            ]
                        )
                        # print('----------dd------')

                        # print(adata_new.shape)
                        # print(adata_new.obsm['X_umap'])
                        adata_new.obsm["clone_vector_emb"] = np.vstack(
                            [
                                adata_new.obsm["X_emb"][i + 1]
                                - adata_new.obsm["X_emb"][i]
                                for i in range(
                                    adata_new.obsm["X_emb"].shape[0] - 1
                                )
                            ]
                            + [np.zeros(2)]
                        )
                        # print('----------ee------')
                        # print(adata_new.obsm['clone_vector_emb'])
                    else:
                        # print('pass-------')
                        continue

                else:
                    time2 = np.where(
                        (adata.obs.time == t) & (adata.obs.clonei == 1)
                    )[0]
                    time4 = np.where(
                        (adata.obs.time == 4) & (adata.obs.clonei == 1)
                    )[0]
                    time6 = np.where(
                        (adata.obs.time == 6) & (adata.obs.clonei == 1)
                    )[0]
                    adata_new = anndata.AnnData(
                        np.vstack(
                            [
                                adata[time2].X.toarray().mean(axis=0),
                                adata[time4].X.toarray().mean(axis=0),
                                adata[time6].X.toarray().mean(axis=0),
                            ]
                        ),
                        layers={
                            "spliced": np.vstack(
                                [
                                    adata[time2]
                                    .layers["spliced"]
                                    .toarray()
                                    .mean(axis=0),
                                    adata[time4]
                                    .layers["spliced"]
                                    .toarray()
                                    .mean(axis=0),
                                    adata[time6]
                                    .layers["spliced"]
                                    .toarray()
                                    .mean(axis=0),
                                ]
                            ),
                            "unspliced": np.vstack(
                                [
                                    adata[time2]
                                    .layers["unspliced"]
                                    .toarray()
                                    .mean(axis=0),
                                    adata[time4]
                                    .layers["unspliced"]
                                    .toarray()
                                    .mean(axis=0),
                                    adata[time6]
                                    .layers["unspliced"]
                                    .toarray()
                                    .mean(axis=0),
                                ]
                            ),
                        },
                        var=adata.var,
                    )

                    print(adata_new.X.sum(axis=1))
                    adata_new.obs.loc[:, "time"] = [2, 4, 6]
                    adata_new.obs.loc[:, "Cell type annotation"] = "Centroid"
                    if not global_traj:
                        adata_new.obs.loc[:, "clonetype"] = (
                            adata[time6].obs.clonetype.unique()
                        )  # use cell fate from last time point
                    adata_new.obs.loc[:, "clones"] = j

                    if "noWell" in adata[time6].obs.columns:
                        adata_new.obs.loc[:, "Well"] = adata[
                            time6
                        ].obs.Well.unique()

                    adata_new.obsm["X_umap"] = np.vstack(
                        [
                            adata[time2].obsm["X_umap"].mean(axis=0),
                            adata[time4].obsm["X_umap"].mean(axis=0),
                            adata[time6].obsm["X_umap"].mean(axis=0),
                        ]
                    )
                    adata_new.obsm["velocity_umap"] = np.vstack(
                        [
                            adata.obsm["X_umap"][time4].mean(axis=0)
                            - adata.obsm["X_umap"][time2].mean(axis=0),
                            adata.obsm["X_umap"][time6].mean(axis=0)
                            - adata.obsm["X_umap"][time4].mean(axis=0),
                            np.zeros(2),
                        ]
                    )

                    # print(adata_new.obsm['velocity_umap'])
                    clone_new = anndata.AnnData(
                        np.vstack(
                            [
                                clone_adata[time2].X.toarray().mean(axis=0),
                                clone_adata[time4].X.toarray().mean(axis=0),
                                clone_adata[time6].X.toarray().mean(axis=0),
                            ]
                        ),
                        obs=adata_new.obs,
                    )
                    clone_new.var_names = clone_adata.var_names
                    clone_new.var = clone_adata.var
                    cen_clones.append(clone_new)
                centroids.append(adata_new)
        print(adata.shape)
        print(len(centroids))
        adata_new = adata.concatenate(
            centroids[0].concatenate(centroids[1:]), join="outer"
        )
        return adata_new


def align_trajectory_diff(
    adatas,
    velocity_embeds,
    density=0.3,
    smooth=0.5,
    input_grid=None,
    input_scale=None,
    min_mass=1.0,
    embed="umap",
    autoscale=False,
    length_cutoff=10,
):
    from scipy.stats import norm as normal
    from scvelo.tools.velocity_embedding import quiver_autoscale
    from sklearn.neighbors import NearestNeighbors

    if input_grid is None and input_scale is None:
        grs = []
        # align embedding points into shared grid across adata
        X_emb = np.vstack([a.obsm[f"X_{embed}"] for a in adatas])
        for dim_i in range(2):
            m, M = np.min(X_emb[:, dim_i]), np.max(X_emb[:, dim_i])
            m = m - 0.01 * np.abs(M - m)
            M = M + 0.01 * np.abs(M - m)
            gr = np.linspace(m, M, int(50 * density))
            grs.append(gr)

        meshes_tuple = np.meshgrid(*grs)
        scale = np.mean([(g[1] - g[0]) for g in grs]) * smooth
        X_grid = np.vstack([i.flat for i in meshes_tuple]).T
    else:
        scale = input_scale
        X_grid = input_grid

    n_neighbors = int(max([a.shape[0] for a in adatas]) / 50)

    results = [X_grid]
    p_mass_list = []
    for adata, velocity_embed in zip(adatas, velocity_embeds):
        nn = NearestNeighbors(n_neighbors=n_neighbors, n_jobs=-1)
        nn.fit(adata.obsm[f"X_{embed}"])
        dists, neighs = nn.kneighbors(X_grid)
        weight = normal.pdf(x=dists, scale=scale)
        # how many cells around a grid points
        p_mass = weight.sum(1)
        V_grid = (velocity_embed[neighs] * weight[:, :, None]).sum(
            1
        ) / np.maximum(1, p_mass)[:, None]
        if autoscale:
            V_grid /= 3 * quiver_autoscale(X_grid, V_grid)
        results.append(V_grid)
        p_mass_list.append(p_mass)

    from functools import reduce

    if input_grid is None and input_scale is None:
        min_mass *= np.percentile(np.hstack(p_mass_list), 99) / 100
        mass_index = reduce(
            np.intersect1d,
            [np.where(p_mass > min_mass)[0] for p_mass in p_mass_list],
        )

    results = np.hstack(results)
    results = results[mass_index]
    print(results.shape)
    length_filter = np.sqrt((results[:, 2:4] ** 2).sum(1)) > length_cutoff
    return results[length_filter]



import os
import pickle
from pathlib import Path
from typing import Any

import zstandard as zstd


__all__ = ["CompressedPickle"]


class CompressedPickle:
    """
    A class for reading and writing zstandard-compressed pickle files.

    Examples:
    >>> import pandas as pd
    >>> test_data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    >>> CompressedPickle.save('test_data.pkl.zst', test_data)
    >>> loaded_data = CompressedPickle.load('test_data.pkl.zst')
    >>> loaded_data.equals(test_data)
    True
    """

    @staticmethod
    def save(file_path: os.PathLike | str, obj: Any) -> None:
        """
        Save the given object to a zstandard-compressed pickle file.

        Args:
            file_path (str): The path of the file to save the object to.
            obj (object): The object to be saved.

        Examples:
        >>> import pandas as pd
        >>> test_data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
        >>> CompressedPickle.save('test_data.pkl.zst', test_data)
        """
        file_path = Path(file_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)

        with file_path.open("wb") as f:
            compression_context = zstd.ZstdCompressor(level=3)
            with compression_context.stream_writer(f) as compressor:
                pickle.dump(obj, compressor)

    @staticmethod
    def load(file_path: os.PathLike | str) -> Any:
        """
        Load an object from a zstandard-compressed pickle file.

        Args:
            file_path (str): The path of the file to load the object from.

        Returns:
            object: The loaded object.

        Examples:
        >>> import pandas as pd
        >>> test_data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
        >>> CompressedPickle.save('test_data.pkl.zst', test_data)
        >>> loaded_data = CompressedPickle.load('test_data.pkl.zst')
        >>> loaded_data.equals(test_data)
        True
        """
        with open(file_path, "rb") as f:
            decompression_context = zstd.ZstdDecompressor()
            with decompression_context.stream_reader(f) as decompressor:
                obj = pickle.load(decompressor)
        return obj



from pathlib import Path

import anndata
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scvelo as scv
import seaborn as sns
from astropy import units as u
from astropy.stats import circstd as acircstd

# from statannotations.Annotator import Annotator
from pyrovelocity.logging import configure_logging


__all__ = [
    "cluster_violin_plots",
    "get_posterior_sample_angle_uncertainty",
    "plot_state_uncertainty",
]

logger = configure_logging(__name__)


def cluster_violin_plots(
    data_model: str,
    adata: anndata.AnnData,
    posterior_samples,
    cluster_key: str,
    violin_flag: bool,
    pairs: list,
    show_outlier: bool,
    fig_name: str | Path = None,
) -> None:
    """Construct violin plots for each cluster."""

    time_cov_list = []
    mag_cov_list = []
    umap_mag_cov_list = []
    umap_angle_std_list = []
    pca_angle_std_list = []
    pca_mag_cov_list = []
    pca_angle_uncertain_list = []
    umap_angle_uncertain_list = []
    names = []
    # print(adata)
    # print(adata.obs[cluster_key])
    fig_name = str(fig_name)

    # get cluster order
    cluster_time_list = []
    clusters = adata.obs[cluster_key].values.categories
    for cluster in clusters:
        adata_cluster = adata[adata.obs[cluster_key] == cluster]
        cluster_time = adata_cluster.obs["velocity_pseudotime"].mean()
        cluster_time_list.append(cluster_time)
    # print(cluster_time_list)
    sorted_cluster_id = sorted(
        range(len(cluster_time_list)),
        key=lambda k: cluster_time_list[k],
        reverse=False,
    )
    order = [str(category) for category in clusters[sorted_cluster_id]]

    umap_cell_angles = posterior_samples["embeds_angle"] / np.pi * 180
    # umap_cell_cirsvar = circvar(umap_cell_angles, axis=0)
    umap_angle_std = acircstd(
        umap_cell_angles * u.deg, method="angular", axis=0
    )
    umap_angle_std_list.append(umap_angle_std)
    umap_angle_uncertain = get_posterior_sample_angle_uncertainty(
        umap_cell_angles
    )
    umap_angle_uncertain_list.append(umap_angle_uncertain)

    pca_cell_vector = posterior_samples["pca_vector_field_posterior_samples"]
    pca_cell_magnitudes = np.sqrt((pca_cell_vector**2).sum(axis=-1))
    pca_cell_magnitudes_mean = pca_cell_magnitudes.mean(axis=-2)
    pca_cell_magnitudes_std = pca_cell_magnitudes.std(axis=-2)
    pca_cell_magnitudes_cov = pca_cell_magnitudes_std / pca_cell_magnitudes_mean
    pca_mag_cov_list.append(pca_cell_magnitudes_cov)

    pca_cell_angles = posterior_samples["pca_embeds_angle"] / np.pi * 180
    # pca_cell_cirsvar = circvar(pca_cell_angles, axis=0)
    pca_cell_cirsstd = acircstd(
        pca_cell_angles * u.deg, method="angular", axis=0
    )
    pca_angle_std_list.append(pca_cell_cirsstd)
    pca_angle_uncertain = get_posterior_sample_angle_uncertainty(
        pca_cell_angles
    )
    pca_angle_uncertain_list.append(pca_angle_uncertain)

    umap_cell_magnitudes = np.sqrt(
        (posterior_samples["vector_field_posterior_samples"] ** 2).sum(axis=-1)
    )
    umap_cell_magnitudes_mean = umap_cell_magnitudes.mean(axis=-2)
    umap_cell_magnitudes_std = umap_cell_magnitudes.std(axis=-2)
    umap_cell_magnitudes_cov = (
        umap_cell_magnitudes_std / umap_cell_magnitudes_mean
    )

    # print(posterior_samples.keys())
    cell_magnitudes = posterior_samples["original_spaces_embeds_magnitude"]
    cell_magnitudes_mean = cell_magnitudes.mean(axis=-2)
    cell_magnitudes_std = cell_magnitudes.std(axis=-2)
    cell_magnitudes_cov = cell_magnitudes_std / cell_magnitudes_mean

    cell_time_mean = posterior_samples["cell_time"].mean(0).flatten()
    cell_time_std = posterior_samples["cell_time"].std(0).flatten()
    cell_time_cov = cell_time_std / cell_time_mean
    time_cov_list.append(cell_time_cov)
    mag_cov_list.append(cell_magnitudes_cov)
    umap_mag_cov_list.append(umap_cell_magnitudes_cov)
    name = list(adata.obs[cluster_key])
    names += name

    can_cast_all = all(name.isdigit() for name in names)

    if can_cast_all:
        processed_names = [f"c_{int(name)}" for name in names]
        order = [f"c_{int_name}" for int_name in order]
        logger.warning(
            "Converted integer index to string 'c_int' for cluster names"
        )
    else:
        processed_names = names

    # print(posterior_samples["pca_vector_field_posterior_samples"].shape)
    # print(posterior_samples["embeds_angle"].shape)
    time_cov_list = np.hstack(time_cov_list)
    mag_cov_list = np.hstack(mag_cov_list)
    pca_mag_cov_list = np.hstack(pca_mag_cov_list)
    pca_angle_std_list = np.hstack(pca_angle_std_list)
    umap_mag_cov_list = np.hstack(umap_mag_cov_list)
    umap_angle_std_list = np.hstack(umap_angle_std_list)

    metrics_df = pd.DataFrame(
        {
            r"$CoV({\mathrm{time}})$": time_cov_list,
            r"$CoV({\mathrm{magnitude}})$": mag_cov_list,
            r"$Std({\mathrm{angle}}_{pca})$": pca_angle_std_list,
            r"$CoV({\mathrm{magnitude}}_{pca})$": pca_mag_cov_list,
            r"$Std({\mathrm{angle}}_{umap})$": umap_angle_std_list,
            r"$CoV({\mathrm{magnitude}}_{umap})$": umap_mag_cov_list,
            "dataset": processed_names,
        }
    )

    max_values, min_values = {}, {}
    for key in metrics_df.keys()[0:6]:
        key_data = metrics_df[key]
        q1, q3 = np.percentile(key_data, (25, 75))
        max_values[key] = q3 + (q3 - q1) * 1.5
        if key_data.min() >= 0:
            min_values[key] = 0
        else:
            min_values[key] = q1 - (q3 - q1) * 1.5
    # print(max_values)
    # print(min_values)

    if "log" in fig_name:
        log_time_cov_list = np.log(time_cov_list)
        log_mag_cov_list = np.log(mag_cov_list)
        log_umap_mag_cov_list = np.log(umap_mag_cov_list)
        pca_angle_uncertain_list = np.hstack(pca_angle_uncertain_list)
        log_pca_mag_cov_list = np.log(pca_mag_cov_list)
        umap_angle_uncertain_list = np.hstack(umap_angle_uncertain_list)
        metrics_df = pd.DataFrame(
            {
                r"$\log(CoV({\mathrm{time}}))$": log_time_cov_list,
                r"$\log(CoV({\mathrm{magnitude}}))$": log_mag_cov_list,
                r"$CircStd({\mathrm{angle}}_{pca})$": pca_angle_uncertain_list,
                r"$\log(CoV({\mathrm{magnitude}}_{pca}))$": log_pca_mag_cov_list,
                r"$CircStd({\mathrm{angle}}_{umap})$": umap_angle_uncertain_list,
                r"$\log(CoV({\mathrm{magnitude}}_{umap}))$": log_umap_mag_cov_list,
                "dataset": processed_names,
            }
        )

    logger.debug(metrics_df.head())
    metrics_df["dataset"] = metrics_df["dataset"].astype("category")
    parameters = {"axes.labelsize": 25, "axes.titlesize": 35}
    plt.rcParams.update(parameters)
    fig, ax = plt.subplots(6, 1)
    ax = ax.flatten()
    fig.set_size_inches(20, 60)
    ax[0].set_title(f"{data_model.split('_')[0]}")
    # ax[0].set_ylim(-5.5, -2)

    if violin_flag:
        for i in range(6):
            sns.violinplot(
                x="dataset",
                y=metrics_df.keys()[i],
                data=metrics_df,
                ax=ax[i],
                order=order,
                # showfliers=show_outlier,
            )
    else:
        for i in range(6):
            sns.boxenplot(
                x="dataset",
                y=metrics_df.keys()[i],
                data=metrics_df,
                ax=ax[i],
                order=order,
                showfliers=show_outlier,
            )

    # if not pairs is None:
    #     for i in range(6):
    #         annotator = Annotator(
    #             ax[i],
    #             pairs,
    #             data=metrics_df,
    #             x="dataset",
    #             y=metrics_df.keys()[i],
    #             order=order,
    #         )
    #         annotator.configure(
    #             test="Mann-Whitney", text_format="star", loc="inside"
    #         )
    #         annotator.apply_and_annotate()

    for axi in ax:
        axi.tick_params(axis="both", labelsize=20)
        axi.set_xlabel("")

    for ext in ["", ".png"]:
        fig.savefig(
            f"{fig_name}{ext}",
            facecolor=fig.get_facecolor(),
            bbox_inches="tight",
            edgecolor="none",
            dpi=300,
        )


def get_posterior_sample_angle_uncertainty(posterior_angles):
    from astropy import units as u
    from astropy.stats import circstd

    x_values = np.arange(360)
    y_values = []
    n_samples = 100
    method = "angular"
    for i in x_values:
        datac = np.linspace(0, i + 1, n_samples) * u.deg
        y_values.append(circstd(datac, method=method))
    y_values = np.array(y_values)
    angle_std = circstd(posterior_angles * u.deg, method="angular", axis=0)
    y_values = y_values.reshape(-1, 1)
    return x_values[np.argmin(np.abs(y_values - angle_std), axis=0)]


def plot_state_uncertainty(
    posterior_samples,
    adata,
    kde=True,
    data="denoised",
    top_percentile=0.9,
    ax=None,
    basis="umap",
):
    if data == "denoised":
        adata.obs["state_uncertain"] = np.sqrt(
            (
                (posterior_samples["st"] - posterior_samples["st"].mean(0)) ** 2
                + (posterior_samples["ut"] - posterior_samples["ut"].mean(0))
                ** 2
            ).sum(-1)
        ).mean(0)
    else:
        adata.obs["state_uncertain"] = np.sqrt(
            (
                (posterior_samples["s"] - posterior_samples["s"].mean(0)) ** 2
                + (posterior_samples["u"] - posterior_samples["u"].mean(0)) ** 2
            ).sum(-1)
        ).mean(0)

    ax = scv.pl.scatter(
        adata,
        basis=basis,
        color="state_uncertain",
        cmap="RdBu_r",
        ax=ax,
        show=False,
        colorbar=True,
        fontsize=7,
    )

    if kde:
        select = adata.obs["state_uncertain"] > np.quantile(
            adata.obs["state_uncertain"], top_percentile
        )
        sns.kdeplot(
            adata.obsm[f"X_{basis}"][:, 0][select],
            adata.obsm[f"X_{basis}"][:, 1][select],
            ax=ax,
            levels=3,
            fill=False,
        )
    else:
        select = None
    return select, ax



import contextlib
import importlib
import inspect
import io
import os
import sys
from inspect import getmembers
from numbers import Integral
from numbers import Real
from pathlib import Path
from pprint import pprint
from types import FunctionType
from types import ModuleType
from typing import Callable
from typing import List
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rich.syntax
import rich.tree
import scvelo as scv
import seaborn as sns
import yaml
from anndata._core.anndata import AnnData
from beartype import beartype
from scvi.data import synthetic_iid

from pyrovelocity.io.compressedpickle import CompressedPickle
from pyrovelocity.logging import configure_logging


# import torch
# from scipy.sparse import issparse
# from sklearn.decomposition import PCA
# from torch.nn.functional import relu
# from pyrovelocity.models._transcription_dynamics import inv
# from pyrovelocity.models import mrna_dynamics

__all__ = [
    "anndata_counts_to_df",
    "attributes",
    "ensure_numpy_array",
    "filter_startswith_dict",
    "generate_public_api",
    "generate_sample_data",
    "internal_help",
    "mae",
    "mae_evaluate",
    "pretty_log_dict",
    "pretty_print_dict",
    "print_anndata",
    "print_attributes",
    "print_config_tree",
    "save_anndata_counts_to_dataframe",
    "str_to_bool",
]

logger = configure_logging(__name__)


def mae(pred_counts, true_counts):
    """
    Computes the mean average error between predicted counts and true counts.

    Args:
        pred_counts (np.ndarray): Predicted counts.
        true_counts (np.ndarray): True counts.

    Returns:
        float: Mean average error.

    Examples:
        >>> import numpy as np
        >>> pred_counts = np.array([[1, 2], [3, 4]])
        >>> true_counts = np.array([[2, 3], [4, 5]])
        >>> mae(pred_counts, true_counts)
        1.0
    """
    error = np.abs(true_counts - pred_counts).sum()
    total = pred_counts.shape[0] * pred_counts.shape[1]
    return (error / total).mean().item()


def ensure_numpy_array(obj):
    return obj.toarray() if hasattr(obj, "toarray") else obj


def mae_evaluate(posterior_samples, adata):
    maes_list = []
    labels = []
    if isinstance(posterior_samples, tuple):
        for model_label, model_obj, split_index in zip(
            ["Poisson train", "Poisson valid"],
            posterior_samples[:2],
            posterior_samples[2:],
        ):
            for sample in range(model_obj["u"].shape[0]):
                maes_list.append(
                    mae(
                        np.hstack(
                            [model_obj["u"][sample], model_obj["s"][sample]]
                        ),
                        np.hstack(
                            [
                                adata.layers["raw_unspliced"].toarray()[
                                    split_index
                                ],
                                adata.layers["raw_spliced"].toarray()[
                                    split_index
                                ],
                            ]
                        ),
                    )
                )
                labels.append(model_label)
    else:
        for model_label, model_obj in zip(
            ["Poisson all cells"], [posterior_samples]
        ):
            for sample in range(model_obj["u"].shape[0]):
                maes_list.append(
                    mae(
                        np.hstack(
                            [model_obj["u"][sample], model_obj["s"][sample]]
                        ),
                        np.hstack(
                            [
                                ensure_numpy_array(
                                    adata.layers["raw_unspliced"]
                                ),
                                ensure_numpy_array(adata.layers["raw_spliced"]),
                            ]
                        ),
                    )
                )
                labels.append(model_label)

    import pandas as pd

    fig, ax = plt.subplots()
    fig.set_size_inches(10, 4)
    df = pd.DataFrame({"MAE": maes_list, "label": labels})
    sns.boxplot(x="label", y="MAE", data=df, ax=ax)
    ax.tick_params(axis="x", rotation=90)
    # print(df.groupby("label").mean())
    return df


def attributes(obj):
    """
    get object attributes
    """
    disallowed_names = {
        name
        for name, value in getmembers(type(obj))
        if isinstance(value, FunctionType)
    }
    return {
        name: getattr(obj, name)
        for name in dir(obj)
        if name[0] != "_"
        and name not in disallowed_names
        and hasattr(obj, name)
    }


def print_attributes(obj):
    """
    print object attributes
    """
    pprint(attributes(obj))


@beartype
def pretty_log_dict(d: dict) -> str:
    dict_as_string = "\n"
    for key, value in d.items():
        # key_colored = colored(key, "green")
        key_colored = key
        value_lines = str(value).split("\n")
        value_colored = "\n".join(
            # colored(line, "white") for line in value_lines
            line
            for line in value_lines
        )
        dict_as_string += f"{key_colored}:\n{value_colored}\n"
    return dict_as_string


@beartype
def pretty_print_dict(d: dict):
    logger.info(pretty_log_dict(d))


@beartype
def print_config_tree(
    config: dict,
    name: str = "",
    theme: str = "nord-darker",
    max_width: int = 148,
):
    config_yaml = yaml.dump(config, default_flow_style=False)
    tree = rich.tree.Tree(name, style="dim", guide_style="dim")
    tree.add(rich.syntax.Syntax(config_yaml, "yaml", theme=theme))

    console = rich.console.Console(width=max_width)
    console.print(tree)


@beartype
def print_docstring(
    obj: Callable | ModuleType,
    theme: str = "nord-darker",
):
    docstring = inspect.getdoc(obj)
    if docstring is None:
        rich.print("[bold red]No docstring found for this object.[/bold red]")
        return
    syntax_highlighted_docstring = rich.syntax.Syntax(
        docstring, "python", theme=theme, line_numbers=False
    )
    rich.print(syntax_highlighted_docstring)


@beartype
def str_to_bool(value: str | bool, default: bool = False) -> bool:
    """
    Convert strings that could be interpreted as booleans to a boolean value,
    with a default fallback.

    Args:
        value (str | bool): input string or boolean value.
        default (bool, optional): Defaults to False.

    Returns:
        bool: boolean interpretation of the input string
    """
    if isinstance(value, bool):
        return value
    if value.lower() in ("true", "t", "1", "yes", "y"):
        return True
    elif value.lower() in ("false", "f", "0", "no", "n"):
        return False
    else:
        return default


def filter_startswith_dict(dictionary_with_underscore_keys):
    """Remove entries from a dictionary whose keys start with an underscore.

    Args:
        dictionary_with_underscore_keys (dict): Dictionary to be filtered.

    Returns:
        dict: Filtered dictionary.
    """
    return {
        k: v
        for k, v in dictionary_with_underscore_keys.items()
        if not k.startswith("_")
    }


def print_anndata(anndata_obj):
    """
    Print a formatted representation of an AnnData object.

    This function produces a custom output for the AnnData object with each
    element of obs, var, uns, obsm, varm, layers, obsp, varp indented and
    displayed on a new line.

    Args:
        anndata_obj (AnnData): The AnnData object to be printed.

    Raises:
        AssertionError: If the input object is not an instance of AnnData.

    Examples:
        >>> import anndata
        >>> import numpy as np
        >>> import pandas as pd
        >>> np.random.seed(42)
        >>> X = np.random.randn(10, 5)
        >>> obs = pd.DataFrame({"clusters_coarse": np.random.randint(0, 2, 10),
        ...                     "clusters": np.random.randint(0, 2, 10),
        ...                     "S_score": np.random.rand(10),
        ...                     "G2M_score": np.random.rand(10)})
        >>> var = pd.DataFrame({"gene_name": [f"gene_{i}" for i in range(5)]})
        >>> adata = AnnData(X, obs=obs, var=var)
        >>> print_anndata(adata)  # doctest: +NORMALIZE_WHITESPACE
    """
    assert isinstance(
        anndata_obj, AnnData
    ), "Input object must be of type AnnData."

    def format_elements(elements):
        formatted = "\n".join([f"        {elem}," for elem in elements])
        return formatted

    anndata_string = [
        f"\nAnnData object with n_obs × n_vars = {anndata_obj.n_obs} × {anndata_obj.n_vars}"
    ]

    properties = {
        "obs": anndata_obj.obs.columns,
        "var": anndata_obj.var.columns,
        "uns": anndata_obj.uns.keys(),
        "obsm": anndata_obj.obsm.keys(),
        "varm": anndata_obj.varm.keys(),
        "layers": anndata_obj.layers.keys(),
        "obsp": anndata_obj.obsp.keys(),
        "varp": anndata_obj.varp.keys(),
    }

    for prop_name, elements in properties.items():
        if len(elements) > 0:
            anndata_string.append(
                f"    {prop_name}:\n{format_elements(elements)}"
            )

    logger.info("\n".join(anndata_string))


def generate_sample_data(
    n_obs: int = 100,
    n_vars: int = 12,
    alpha: float = 5,
    beta: float = 0.5,
    gamma: float = 0.3,
    alpha_: float = 0,
    noise_model: str = "gillespie",
    random_seed: int = 0,
) -> AnnData:
    """
    Generate synthetic single-cell RNA sequencing data with spliced and unspliced layers.
    If using the "iid" noise model, the data will be generated with scvi.data.synthetic_iid.
    If using the "normal" or "gillespie" noise model, the data will be generated with
    scvelo.datasets.simulation accounting for the given expression dynamics parameters.

    Args:
        n_obs (int, optional): Number of observations (cells). Default is 100.
        n_vars (int, optional): Number of variables (genes). Default is 12.
        alpha (float, optional): Transcription rate. Default is 5.
        beta (float, optional): Splicing rate. Default is 0.5.
        gamma (float, optional): Degradation rate. Default is 0.3.
        alpha_ (float, optional): Additional transcription rate. Default is 0.
        noise_model (str, optional): Noise model to be used. Must be one of 'iid', 'gillespie', or 'normal'. Default is 'gillespie'.
        random_seed (int, optional): Random seed for reproducibility. Default is 0.

    Returns:
        AnnData: An AnnData object containing the generated synthetic data.

    Raises:
        ValueError: If noise_model is not one of 'iid', 'gillespie', or 'normal'.

    Examples:
        >>> from pyrovelocity.utils import generate_sample_data, print_anndata
        >>> adata = generate_sample_data(random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(n_obs=50, n_vars=10, noise_model="normal", random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(n_obs=50, n_vars=10, noise_model="iid", random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(noise_model="wishful thinking")
        Traceback (most recent call last):
            ...
        ValueError: noise_model must be one of 'iid', 'gillespie', 'normal'
    """
    if noise_model == "iid":
        adata = synthetic_iid(
            batch_size=n_obs,
            n_genes=n_vars,
            n_batches=1,
            n_labels=1,
        )
        adata.layers["spliced"] = adata.X.copy()
        adata.layers["unspliced"] = adata.X.copy()
    elif noise_model in {"gillespie", "normal"}:
        adata = scv.datasets.simulation(
            random_seed=random_seed,
            n_obs=n_obs,
            n_vars=n_vars,
            alpha=alpha,
            beta=beta,
            gamma=gamma,
            alpha_=alpha_,
            noise_model=noise_model,
        )
    else:
        raise ValueError(
            "noise_model must be one of 'iid', 'gillespie', 'normal'"
        )
    return adata


@beartype
def anndata_counts_to_df(
    adata: AnnData,
) -> Tuple[pd.DataFrame, Integral, Integral, Real, Real]:
    spliced_df = pd.DataFrame(
        ensure_numpy_array(adata.layers["raw_spliced"]),
        index=list(adata.obs_names),
        columns=list(adata.var_names),
    )
    unspliced_df = pd.DataFrame(
        ensure_numpy_array(adata.layers["raw_unspliced"]),
        index=list(adata.obs_names),
        columns=list(adata.var_names),
    )

    spliced_melted = spliced_df.reset_index().melt(
        id_vars="index", var_name="var_name", value_name="spliced"
    )
    unspliced_melted = unspliced_df.reset_index().melt(
        id_vars="index", var_name="var_name", value_name="unspliced"
    )

    df = spliced_melted.merge(unspliced_melted, on=["index", "var_name"])

    df = df.rename(columns={"index": "obs_name"})

    total_obs = adata.n_obs
    total_var = adata.n_vars

    max_spliced = adata.layers["raw_spliced"].max()
    max_unspliced = adata.layers["raw_unspliced"].max()

    logger.warning(
        f"Type of df is {type(df)}\n"
        f"Type of total_obs is {type(total_obs)}\n"
        f"Type of total_var is {type(total_var)}\n"
        f"Type of max_spliced is {type(max_spliced)}\n"
        f"Type of max_unspliced is {type(max_unspliced)}\n"
    )

    return (
        df,
        total_obs,
        total_var,
        max_spliced,
        max_unspliced,
    )


# TODO: migrate to parquet
@beartype
def save_anndata_counts_to_dataframe(
    adata: AnnData,
    dataframe_path: os.PathLike | str,
) -> Path:
    logger.info(f"Saving AnnData object to dataframe: {dataframe_path}")
    df_tuple = anndata_counts_to_df(adata)

    CompressedPickle.save(dataframe_path, df_tuple)

    if not os.path.isfile(dataframe_path) or not os.access(
        dataframe_path, os.R_OK
    ):
        raise FileNotFoundError(f"Failed to create readable {dataframe_path}")

    return Path(dataframe_path)


@beartype
def generate_public_api(module_name: str) -> List[str]:
    """
    Generates a list of names of functions and classes defined in a specified
    module. This may be used, for example, to generate a candidate for the
    public interface for a module or package.

    Args:
        module_name (str): The name of the module.

    Returns:
        List[str]: A list of names of functions and classes defined in the module.
    """
    module = importlib.import_module(module_name)
    public_api: List[str] = [
        name
        for name, obj in inspect.getmembers(
            module,
            lambda obj: (inspect.isfunction(obj) or inspect.isclass(obj))
            and obj.__module__ == module.__name__,
        )
    ]
    return public_api


class SuppressOutput:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def internal_help(obj: Callable | ModuleType):
    """Generate help text for a callable or module.

    Args:
        obj (Callable | ModuleType): The object to generate help text for.
    """
    captured_output = io.StringIO()
    with contextlib.redirect_stdout(captured_output):
        help(obj)

    help_text = captured_output.getvalue()
    lines = help_text.split("\n")

    processed_lines = []
    for line in lines:
        if line.startswith("FILE"):
            break
        processed_lines.append(line)

    print("\n".join(processed_lines))


# TODO: remove unused functions
# def log(x):
#     """
#     Computes the element-wise natural logarithm of a tensor, while clipping
#     its values to avoid numerical instability.

#     Args:
#         x (torch.Tensor): Input tensor.

#     Returns:
#         torch.Tensor: Tensor with element-wise natural logarithm of x.

#     Examples:
#         >>> import torch
#         >>> x = torch.tensor([0.0001, 0.5, 0.9999])
#         >>> log(x)
#         tensor([-9.2103e+00, -6.9315e-01, -1.0002e-04])
#     """
#     eps = torch.finfo(x.dtype).eps
#     return torch.log(x.clamp(eps, 1 - eps))


# TODO: remove unused functions
# def tau_inv(
#     u=None, s=None, u0=None, s0=None, alpha=None, beta=None, gamma=None
# ):
#     """
#     Computes the inverse tau given the parameters and initial conditions.

#     Args:
#         u (torch.Tensor): Value of u.
#         s (torch.Tensor): Value of s.
#         u0 (torch.Tensor): Initial value of u.
#         s0 (torch.Tensor): Initial value of s.
#         alpha (torch.Tensor): Alpha parameter.
#         beta (torch.Tensor): Beta parameter.
#         gamma (torch.Tensor): Gamma parameter.

#     Returns:
#         torch.Tensor: Inverse tau.

#     Examples:
#         >>> import torch
#         >>> u = torch.tensor(0.6703)
#         >>> s = torch.tensor(0.4596)
#         >>> u0 = torch.tensor(1.0)
#         >>> s0 = torch.tensor(0.5)
#         >>> alpha = torch.tensor(0.5)
#         >>> beta = torch.tensor(0.4)
#         >>> gamma = torch.tensor(0.3)
#         >>> tau_inv(u, s, u0, s0, alpha, beta, gamma)
#         tensor(3.9736e-07)
#     """
#     beta_ = beta * inv(gamma - beta)
#     xinf = alpha / gamma - beta_ * (alpha / beta)
#     tau1 = (
#         -1.0 / gamma * log((s - beta_ * u - xinf) * inv(s0 - beta_ * u0 - xinf))
#     )

#     uinf = alpha / beta
#     tau2 = -1 / beta * log((u - uinf) * inv(u0 - uinf))
#     tau = torch.where(beta > gamma, tau1, tau2)
#     return relu(tau)


# TODO: remove unused functions
# def init_with_all_cells(
#     adata,
#     input_type="knn",
#     shared_time=True,
#     latent_factor="linear",
#     latent_factor_size=10,
#     plate_size=2,
#     num_aux_cells=200,
#     init_smooth=True,
# ):
#     ## hard to use unsmoothed data for initialization
#     ## always initialize the model with smoothed data
#     if "Mu" in adata.layers and "Ms" in adata.layers and init_smooth:
#         u_obs = torch.tensor(adata.layers["Mu"], dtype=torch.float32)
#         s_obs = torch.tensor(adata.layers["Ms"], dtype=torch.float32)
#     elif "spliced" in adata.layers:
#         u_obs = torch.tensor(
#             adata.layers["unspliced"].toarray()
#             if issparse(adata.layers["unspliced"])
#             else adata.layers["unspliced"],
#             dtype=torch.float32,
#         )
#         s_obs = torch.tensor(
#             adata.layers["spliced"].toarray()
#             if issparse(adata.layers["spliced"])
#             else adata.layers["spliced"],
#             dtype=torch.float32,
#         )
#     else:
#         raise

#     ub_u = torch.stack(
#         [torch.quantile(u[u > 0], 0.99) for u in torch.unbind(u_obs, axis=1)]
#     )
#     ub_s = torch.stack(
#         [torch.quantile(s[s > 0], 0.99) for s in torch.unbind(s_obs, axis=1)]
#     )
#     s_mask = (s_obs > 0) & (s_obs <= ub_s)
#     u_mask = (u_obs > 0) & (u_obs <= ub_u)
#     # include zeros
#     training_mask = s_mask & u_mask

#     u_scale = torch.stack(
#         [u[u > 0].std() for u in torch.unbind(u_obs * training_mask, axis=1)]
#     )
#     s_scale = torch.stack(
#         [s[s > 0].std() for s in torch.unbind(s_obs * training_mask, axis=1)]
#     )
#     scale = u_scale / s_scale

#     lb_steady_u = torch.stack(
#         [
#             torch.quantile(u[u > 0], 0.98)
#             for u in torch.unbind(u_obs * training_mask, axis=1)
#         ]
#     )
#     lb_steady_s = torch.stack(
#         [
#             torch.quantile(s[s > 0], 0.98)
#             for s in torch.unbind(s_obs * training_mask, axis=1)
#         ]
#     )

#     steady_u_mask = training_mask & (u_obs >= lb_steady_u)
#     steady_s_mask = training_mask & (s_obs >= lb_steady_s)

#     u_obs = u_obs / scale
#     u_inf = (u_obs * (steady_u_mask | steady_s_mask)).sum(dim=0) / (
#         steady_u_mask | steady_s_mask
#     ).sum(dim=0)
#     s_inf = (s_obs * steady_s_mask).sum(dim=0) / steady_s_mask.sum(dim=0)

#     gamma = (u_obs * steady_s_mask * s_obs).sum(axis=0) / (
#         (steady_s_mask * s_obs) ** 2
#     ).sum(axis=0) + 1e-6
#     gamma = torch.where(gamma < 0.05 / scale, gamma * 1.2, gamma)
#     gamma = torch.where(gamma > 1.5 / scale, gamma / 1.2, gamma)
#     alpha = gamma * s_inf
#     beta = alpha / u_inf

#     switching = tau_inv(u_inf, s_inf, 0.0, 0.0, alpha, beta, gamma)
#     tau = tau_inv(u_obs, s_obs, 0.0, 0.0, alpha, beta, gamma)
#     tau = torch.where(tau >= switching, switching, tau)
#     tau_ = tau_inv(u_obs, s_obs, u_inf, s_inf, 0.0, beta, gamma)
#     tau_ = torch.where(
#         tau_ >= tau_[s_obs > 0].max(dim=0)[0],
#         tau_[s_obs > 0].max(dim=0)[0],
#         tau_,
#     )
#     ut, st = mrna_dynamics(tau, 0.0, 0.0, alpha, beta, gamma)
#     ut_, st_ = mrna_dynamics(tau_, u_inf, s_inf, 0.0, beta, gamma)

#     u_scale_ = u_scale / scale
#     state_on = ((ut - u_obs) / u_scale_) ** 2 + ((st - s_obs) / s_scale) ** 2
#     state_off = ((ut_ - u_obs) / u_scale_) ** 2 + ((st_ - s_obs) / s_scale) ** 2
#     cell_gene_state_logits = state_on - state_off
#     cell_gene_state = cell_gene_state_logits < 0
#     t = torch.where(cell_gene_state_logits < 0, tau, tau_ + switching)
#     init_values = {
#         "alpha": alpha,
#         "beta": beta,
#         "gamma": gamma,
#         "switching": switching,
#         "latent_time": t,
#         "u_scale": u_scale,
#         "s_scale": s_scale,
#         "u_inf": u_inf,
#         "s_inf": s_inf,
#     }
#     if input_type == "knn":
#         init_values["mask"] = training_mask

#     init_values["cell_gene_state"] = cell_gene_state.int()
#     if latent_factor == "linear":
#         if "spliced" in adata.layers:
#             u_obs = torch.tensor(
#                 adata.layers["unspliced"].toarray()
#                 if issparse(adata.layers["unspliced"])
#                 else adata.layers["unspliced"],
#                 dtype=torch.float32,
#             )
#             s_obs = torch.tensor(
#                 adata.layers["spliced"].toarray()
#                 if issparse(adata.layers["spliced"])
#                 else adata.layers["spliced"],
#                 dtype=torch.float32,
#             )
#         test = np.hstack([u_obs, s_obs])
#         pca = PCA(n_components=latent_factor_size)
#         pca.fit(test)
#         X_train_pca = pca.transform(test)
#         init_values["cell_codebook"] = torch.tensor(pca.components_)
#         init_values["u_pcs_mean"] = torch.tensor(pca.mean_[: adata.shape[1]])
#         init_values["s_pcs_mean"] = torch.tensor(pca.mean_[adata.shape[1] :])
#         if plate_size == 2:
#             init_values["cell_code"] = (
#                 torch.tensor(X_train_pca).unsqueeze(-1).transpose(-1, -2)
#             )
#         else:
#             init_values["cell_code"] = torch.tensor(X_train_pca)

#     if num_aux_cells > 0:
#         np.random.seed(99)
#         if "cytotrace" in adata.obs.columns:
#             order_aux = np.array_split(
#                 np.sort(adata.obs["cytotrace"].values), 50
#             )
#             order_aux_list = []
#             for i in order_aux:
#                 order_aux_list.append(
#                     np.random.choice(
#                         np.where(
#                             (adata.obs["cytotrace"].values > i[0])
#                             & (adata.obs["cytotrace"].values < i[-1])
#                         )[0]
#                     )
#                 )
#             order_aux = np.array(order_aux_list)
#         else:
#             order_aux = np.argsort(
#                 (adata.layers["spliced"].toarray() > 0).sum(axis=1)
#             )[::-1]

#         init_values["order_aux"] = torch.from_numpy(
#             order_aux[:num_aux_cells].copy()
#         )

#         if input_type == "raw":
#             u_obs = adata.layers["raw_unspliced"].toarray()
#             s_obs = adata.layers["raw_spliced"].toarray()
#         init_values["aux_u_obs"] = torch.tensor(
#             u_obs[order_aux][:num_aux_cells]
#         )
#         init_values["aux_s_obs"] = torch.tensor(
#             s_obs[order_aux][:num_aux_cells]
#         )
#         init_values["cell_gene_state_aux"] = cell_gene_state[:num_aux_cells]
#         init_values["latent_time_aux"] = t[:num_aux_cells]

#         if latent_factor == "linear":
#             init_values["cell_code_aux"] = (
#                 torch.tensor(X_train_pca)
#                 .unsqueeze(-1)
#                 .transpose(-1, -2)[:num_aux_cells]
#             )

#     if shared_time:
#         cell_time = t.mean(dim=-1, keepdims=True)
#         init_values["cell_time"] = cell_time
#         init_values["latent_time"] = init_values["latent_time"] - cell_time

#         if num_aux_cells > 0:
#             init_values["cell_time_aux"] = cell_time[:num_aux_cells]
#             init_values["latent_time_aux"] = (
#                 init_values["latent_time_aux"] - init_values["cell_time_aux"]
#             )

#     for key in init_values:
#         print(key, init_values[key].shape)
#         print(init_values[key].isnan().sum())
#         assert init_values[key].isnan().sum() == 0
#     return init_values


# TODO: remove unused functions
# def trace(func):
#     def wrapper(*args, **kwargs):
#         params = list(args) + [f"{k}={v}" for k, v in kwargs.items()]
#         params_str = ",\n    ".join(map(str, params))
#         print(f"{func.__name__}(\n    {params_str},\n)")
#         return func(*args, **kwargs)

#     return wrapper


# def velocity_dus_dt(alpha, beta, gamma, tau, x):
#     """
#     Computes the velocity du/dt and ds/dt.

#     Args:
#         alpha (torch.Tensor): Alpha parameter.
#         beta (torch.Tensor): Beta parameter.
#         gamma (torch.Tensor): Gamma parameter.
#         tau (torch.Tensor): Time points.
#         x (Tuple[torch.Tensor, torch.Tensor]): Tuple containing u and s.

#     Returns:
#         Tuple[torch.Tensor, torch.Tensor]: Tuple containing du/dt and ds/dt.

#     Examples:
#         >>> import torch
#         >>> alpha = torch.tensor(0.5)
#         >>> beta = torch.tensor(0.4)
#         >>> gamma = torch.tensor(0.3)
#         >>> tau = torch.tensor(2.0)
#         >>> x = (torch.tensor(1.0), torch.tensor(0.5))
#         >>> velocity_dus_dt(alpha, beta, gamma, tau, x)
#         (tensor(0.1000), tensor(0.2500))
#     """
#     u, s = x
#     du_dt = alpha - beta * u
#     ds_dt = beta * u - gamma * s
#     return du_dt, ds_dt


# def rescale_time(dx_dt, t_start, t_end):
#     """
#     Converts an ODE to be solved on a batch of different time intervals into an
#     equivalent system of ODEs to be solved on [0, 1].

#     Args:
#         dx_dt (Callable): Function representing the ODE system.
#         t_start (torch.Tensor): Start time of the time interval.
#         t_end (torch.Tensor): End time of the time interval.

#     Returns:
#         Callable: Function representing the rescaled ODE system.

#     Examples:
#         >>> import torch
#         >>> def dx_dt(t, x):
#         ...     return -x
#         >>> t_start = torch.tensor(0.0)
#         >>> t_end = torch.tensor(1.0)
#         >>> rescaled_dx_dt = rescale_time(dx_dt, t_start, t_end)
#         >>> rescaled_dx_dt(torch.tensor(0.5), torch.tensor(2.0))
#         tensor(-2.)
#     """
#     dt = t_end - t_start

#     @functools.wraps(dx_dt)
#     def dx_ds(s, *args):
#         # move any batch dimensions in s to the left of all batch dimensions in dt
#         # s_shape = (1,) if len(s.shape) == 0 else s.shape  # XXX unnecessary?
#         s = s.reshape(s.shape + (1,) * len(dt.shape))
#         t = s * dt + t_start
#         xts = dx_dt(t, *args)
#         if isinstance(xts, tuple):
#             xss = tuple(xt * dt for xt in xts)
#         else:
#             xss = xts * dt
#         return xss

#     return dx_ds


# def ode_mrna_dynamics(tau, u0, s0, alpha, beta, gamma):
#     """
#     Solves the ODE system for mRNA dynamics.

#     Args:
#         tau (torch.Tensor): Time points.
#         u0 (torch.Tensor): Initial value of u.
#         s0 (torch.Tensor): Initial value of s.
#         alpha (torch.Tensor): Alpha parameter.
#         beta (torch.Tensor): Beta parameter.
#         gamma (torch.Tensor): Gamma parameter.

#     Returns:
#         Tuple[torch.Tensor, torch.Tensor]: Tuple containing the final values of u and s.
#     """
#     """
#     Examples:
#         >>> import torch
#         >>> tau = torch.tensor(2.0)
#         >>> u0 = torch.tensor(1.0)
#         >>> s0 = torch.tensor(0.5)
#         >>> alpha = torch.tensor(0.5)
#         >>> beta = torch.tensor(0.4)
#         >>> gamma = torch.tensor(0.3)
#         >>> ode_mrna_dynamics(tau, u0, s0, alpha, beta, gamma)
#         (tensor(0.6703), tensor(0.4596))
#     """
#     dx_dt = functools.partial(velocity_dus_dt, alpha, beta, gamma)
#     dx_dt = rescale_time(
#         dx_dt, torch.tensor(0.0, dtype=tau.dtype, device=tau.device), tau
#     )
#     grid = torch.linspace(0.0, 1.0, 100, dtype=tau.dtype, device=tau.device)
#     x0 = (u0.expand_as(tau), s0.expand_as(tau))
#     # xts = odeint_adjoint(dx_dt, x0, grid, adjoint_params=(alpha, beta, gamma, tau))
#     xts = odeint(dx_dt, x0, grid)
#     uts, sts = xts
#     ut, st = uts[-1], sts[-1]
#     return ut, st


# def get_velocity_samples(posterior_samples, model):
#     """
#     Computes the velocity samples from the posterior samples.

#     Args:
#         posterior_samples (dict): Dictionary containing posterior samples.
#         model: Model used for predictions.

#     Returns:
#         torch.Tensor: Velocity samples.

#     Examples:
#         >>> import torch
#         >>> posterior_samples = {
#         ...     "beta": torch.tensor([[0.4]]),
#         ...     "gamma": torch.tensor([[0.3]]),
#         ...     "u_scale": torch.tensor([[[1.0, 2.0]]]),
#         ...     "s_scale": torch.tensor([[[2.0, 4.0]]]),
#         ...     "u": torch.tensor([[1.0, 2.0]]),
#         ...     "s": torch.tensor([[0.5, 1.0]])
#         ... }
#         >>> model = None  # Model is not used in the function
#         >>> get_velocity_samples(posterior_samples, model)
#         tensor([[0.6500, 1.3000]])
#     """
#     beta = posterior_samples["beta"].mean(0)[0]
#     gamma = posterior_samples["gamma"].mean(0)[0]
#     scale = (
#         posterior_samples["u_scale"][:, 0, :]
#         / posterior_samples["s_scale"][:, 0, :]
#     ).mean(0)
#     ut = posterior_samples["u"] / scale
#     st = posterior_samples["s"]
#     v = beta * ut - gamma * st
#     return v

# def debug(x):
#     if torch.any(torch.isnan(x)):
#         print("nan number: ", torch.isnan(x).sum())
#         pdb.set_trace()


# def site_is_discrete(site: dict) -> bool:
#     return (
#         site["type"] == "sample"
#         and not site["is_observed"]
#         and getattr(site["fn"], "has_enumerate_support", False)
#     )


# def get_pylogger(name=__name__, log_level="DEBUG") -> logging.Logger:
#     """Initializes multi-GPU-friendly python command line logger."""

#     formatter = colorlog.ColoredFormatter(
#         "%(log_color)s%(levelname)s:%(name)s: %(message)s"
#         # "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
#         # datefmt=None,
#         # reset=True,
#         # log_colors={
#         #     "debug": "cyan",
#         #     "info": "green",
#         #     "warning": "yellow",
#         #     "error": "red",
#         #     "exception": "red",
#         #     "fatal": "red",
#         #     "critical": "red",
#         #     },
#     )

#     handler = colorlog.StreamHandler()
#     handler.setFormatter(formatter)
#     logger = colorlog.getLogger(name)
#     logger.setLevel(log_level)
#     logger.addHandler(handler)
#     logger.propagate = False

#     # this ensures all logging levels get marked with the rank zero decorator
#     # otherwise logs would get multiplied for each GPU process in multi-GPU setup
#     logging_levels = (
#         "debug",
#         "info",
#         "warning",
#         "error",
#         "exception",
#         "fatal",
#         "critical",
#     )
#     for level in logging_levels:
#         setattr(logger, level, rank_zero_only(getattr(logger, level)))

#     return logger
#     for level in logging_levels:
#         setattr(logger, level, rank_zero_only(getattr(logger, level)))

#     return logger



import pyrovelocity.tasks.data
import pyrovelocity.tasks.postprocess
import pyrovelocity.tasks.preprocess
import pyrovelocity.tasks.summarize
import pyrovelocity.tasks.train


__all__ = [
    "data",
    "postprocess",
    "preprocess",
    "summarize",
    "train",
]



import shutil
import tempfile
from pathlib import Path
from typing import Tuple

import nbformat
import papermill as pm
import rich_click as click
from flytekit import task
from flytekit import workflow
from jupytext import read
from jupytext import write
from nbconvert import HTMLExporter
from nbconvert import PDFExporter
from nbconvert.writers import FilesWriter

from pyrovelocity.logging import configure_logging


__all__ = ["notebook_workflow"]

logger = configure_logging(__name__)

click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.USE_MARKDOWN = True

CACHE_FLAG = True
CACHE_VERSION = "0.1.dev0"


@task(cache=CACHE_FLAG, cache_version=CACHE_VERSION)
def convert_py_to_ipynb(py_file_path: str) -> str:
    if CACHE_FLAG:
        logger.warning(
            f"\nCache flag enabled: {CACHE_FLAG}\n"
            f"Cache version: {CACHE_VERSION}\n"
        )

    logger.info(f"Converting {py_file_path} to ipynb format.")

    notebook = read(py_file_path)
    ipynb_file_path = py_file_path.replace(".py", ".ipynb")
    write(notebook, ipynb_file_path)

    if not Path(ipynb_file_path).exists():
        raise FileNotFoundError(
            "Conversion failed or output file does not exist."
        )
    else:
        logger.info(f"Conversion successful. Output: {ipynb_file_path}")

    return ipynb_file_path


@task(cache=CACHE_FLAG, cache_version=CACHE_VERSION)
def execute_notebook(ipynb_file_path: str, test_mode: bool) -> str:
    executed_ipynb_file_path = ipynb_file_path.replace(
        ".ipynb", "_executed.ipynb"
    )

    logger.info(
        f"Executing {ipynb_file_path} and saving to {executed_ipynb_file_path}."
    )
    if test_mode:
        logger.warning(f"Test mode enabled: {test_mode}")

    pm.execute_notebook(
        ipynb_file_path,
        executed_ipynb_file_path,
        parameters=dict(TEST_MODE=test_mode),
    )

    executed_ipynb_file = Path(executed_ipynb_file_path)
    if not executed_ipynb_file.exists():
        raise FileNotFoundError(
            "Execution failed or output file {executed_ipynb_file_path} does not exist."
        )
    else:
        logger.info(f"Execution successful. Output: {executed_ipynb_file_path}")

    return executed_ipynb_file_path


@task(cache=CACHE_FLAG, cache_version=CACHE_VERSION)
def convert_ipynb_to_html_and_pdf(
    executed_ipynb_file_path: str
) -> Tuple[str, str]:
    logger.info(
        f"Converting {executed_ipynb_file_path} to html and pdf formats."
    )
    notebook_node = nbformat.read(executed_ipynb_file_path, as_version=4)

    html_exporter = HTMLExporter()
    pdf_exporter = PDFExporter()

    (body_html, resources_html) = html_exporter.from_notebook_node(
        notebook_node
    )
    del resources_html
    html_file_path = executed_ipynb_file_path.replace(".ipynb", ".html")
    with open(html_file_path, "w") as html_file:
        html_file.write(body_html)

    with tempfile.TemporaryDirectory() as tmpdir:
        pdf_exporter = PDFExporter()
        pdf_exporter.output_files_dir = tmpdir
        (body_pdf, resources_pdf) = pdf_exporter.from_notebook_node(
            notebook_node
        )

        pdf_file_name = Path(executed_ipynb_file_path).stem + ".pdf"
        temp_pdf_file_path = Path(tmpdir) / pdf_file_name

        pdf_writer = FilesWriter(build_directory=str(tmpdir))
        pdf_writer.write(
            body_pdf, resources_pdf, notebook_name=temp_pdf_file_path.stem
        )

        final_pdf_file_path = Path(
            executed_ipynb_file_path.replace(".ipynb", ".pdf")
        )
        shutil.copy(temp_pdf_file_path, final_pdf_file_path)

    if not Path(html_file_path).exists():
        raise FileNotFoundError(
            f"Conversion failed or output file {html_file_path} does not exist."
        )
    if not final_pdf_file_path.exists():
        raise FileNotFoundError(
            "Conversion failed or output file {final_pdf_file_path} does not exist."
        )
    logger.info(
        f"Conversion successful. Outputs:\n"
        f"HTML File: {html_file_path}\n"
        f"PDF File: {final_pdf_file_path}\n"
    )

    return html_file_path, str(final_pdf_file_path)


@workflow
def notebook_workflow(
    py_file_path: str = "notebook.py", test_mode: bool = True
) -> Tuple[str, str, str]:
    """
    This workflow converts a python percent-formatted notebook script into an
    executed ipython notebook and derived html and pdf formats. This can be
    tested locally with a command similar to:

    ```bash
    pyrovelocity \
        execution_context=local_shell \
        entity_config=notebooks_notebook_workflow \
        entity_config.inputs._args_.0.py_file_path=scripts/examples/literate_template.py \
        entity_config.inputs._args_.0.test_mode=true
    ```

    Alternatively, use the `execute` command in the CLI to run the workflow
    similar to:
    
    ```bash
    python -m pyrovelocity.workflows.notebooks execute \
        --py-file-path scripts/examples/literate_template.py \
        --test-mode
    ```

    See `python -m pyrovelocity.workflows.notebooks execute --help` for more
    details.

    Args:
        py_file_path (str, optional): Path to notebook script. Defaults to "notebook.py".
        test_mode (bool, optional): Flag indicating execution in test mode. Defaults to True.

    Returns:
        Tuple[str, str, str]: Paths to executed ipynb, html, and pdf files.
    """
    ipynb_file_path = convert_py_to_ipynb(py_file_path=py_file_path)
    executed_ipynb_file_path = execute_notebook(
        ipynb_file_path=ipynb_file_path, test_mode=test_mode
    )
    html_file_path, pdf_file_path = convert_ipynb_to_html_and_pdf(
        executed_ipynb_file_path=executed_ipynb_file_path
    )
    return executed_ipynb_file_path, html_file_path, pdf_file_path


@click.group(
    invoke_without_command=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.pass_context
def cli(ctx):
    """
    # notebooks workflow
    The _**notebooks workflow**_ converts a python
    [percent formatted](https://jupytext.readthedocs.io/en/latest/formats-scripts.html)
    notebook script into ipython notebook format, executes the notebook with
    papermill and converts it to html and pdf format.

    Pass -h or --help to each command group listed below for detailed help.

    [notebook percent format](https://jupytext.readthedocs.io/en/latest/formats-scripts.html)
    [papermill](https://github.com/nteract/papermill)
    [nbformat](https://github.com/jupyter/nbformat)
    [nbconvert](https://github.com/jupyter/nbconvert)
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@cli.command("execute")
@click.option(
    "--py-file-path",
    type=str,
    help="Path to the percent-formatted .py file (e.g. notebook.py) to be processed",
    required=True,
)
@click.option(
    "--test-mode",
    is_flag=True,
    default=False,
    help="Flag to run the notebook in test mode.",
)
def cli_execute_notebook(py_file_path: str, test_mode: bool):
    """
    # execute
    Execute the notebooks workflow with the provided python file path.

    The python file should be in
    [percent format](https://jupytext.readthedocs.io/en/latest/formats-scripts.html)
    and contain markdown cells for literate programming.
    """
    result = notebook_workflow(py_file_path=py_file_path, test_mode=test_mode)
    click.echo(
        f"Workflow completed with outputs:\n\n"
        f"Executed Notebook: {result[0]}\n"
        f"HTML File: {result[1]}\n"
        f"PDF File: {result[2]}\n"
    )


if __name__ == "__main__":
    cli()



from pyrovelocity.models._deterministic_inference import (
    deterministic_transcription_splicing_probabilistic_model,
)
from pyrovelocity.models._deterministic_simulation import (
    solve_transcription_splicing_model,
)
from pyrovelocity.models._deterministic_simulation import (
    solve_transcription_splicing_model_analytical,
)
from pyrovelocity.models._transcription_dynamics import mrna_dynamics
from pyrovelocity.models._velocity import PyroVelocity


__all__ = [
    deterministic_transcription_splicing_probabilistic_model,
    mrna_dynamics,
    PyroVelocity,
    solve_transcription_splicing_model,
    solve_transcription_splicing_model_analytical,
]



from pyrovelocity.plots._dynamics import (
    plot_deterministic_simulation_phase_portrait,
)
from pyrovelocity.plots._dynamics import (
    plot_deterministic_simulation_trajectories,
)
from pyrovelocity.plots._experimental import plot_t0_selection
from pyrovelocity.plots._genes import plot_gene_ranking
from pyrovelocity.plots._parameters import (
    plot_parameter_posterior_distributions,
)
from pyrovelocity.plots._predictive import (
    extrapolate_prediction_sample_predictive,
)
from pyrovelocity.plots._predictive import posterior_curve
from pyrovelocity.plots._rainbow import rainbowplot
from pyrovelocity.plots._summary import plot_gene_selection_summary
from pyrovelocity.plots._time import plot_posterior_time
from pyrovelocity.plots._time import plot_shared_time_uncertainty
from pyrovelocity.plots._uncertainty import cluster_violin_plots
from pyrovelocity.plots._uncertainty import (
    get_posterior_sample_angle_uncertainty,
)
from pyrovelocity.plots._uncertainty import plot_state_uncertainty
from pyrovelocity.plots._vector_fields import plot_vector_field_summary


__all__ = [
    "cluster_violin_plots",
    "extrapolate_prediction_sample_predictive",
    "get_posterior_sample_angle_uncertainty",
    "plot_deterministic_simulation_phase_portrait",
    "plot_deterministic_simulation_trajectories",
    "plot_gene_ranking",
    "plot_parameter_posterior_distributions",
    "plot_posterior_time",
    "plot_shared_time_uncertainty",
    "plot_state_uncertainty",
    "plot_t0_selection",
    "posterior_curve",
    "rainbowplot",
    "plot_vector_field_summary",
    "plot_gene_selection_summary",
]
